from upydevice import Device, DeviceException
from upydev.commandlib import _CMDDICT_
import sys
from upydev.helpinfo import see_help

KEY_N_ARGS = {'sd_enable': ['po']}

VALS_N_ARGS = ['po']


SD_COMMANDS_HELP = """
> SD: Usage '$ upydev COMMAND [opts]'
    * COMMAND:
        - sd_enable: to enable/disable the LDO 3.3V regulator that powers the SD module
                     use -po option to indicate the Pin.
        - sd_init: to initialize the sd card; (spi must be configured first)
                   create sd object and mounts as a filesystem, needs sdcard.py see
                   https://github.com/Carglglz/upydev/blob/master/DOCS/Documentation.md#sd_init
        - sd_deinit: to unmount sd card
        - sd_auto: experimental command, needs a sd module with sd detection pin
                   and the SD_AM.py script. Enable an Interrupt
                   with the sd detection pin, so it mounts the sd when is detected,
                   and unmount the sd card when is extracted. See more info in:
                   https://github.com/Carglglz/upydev/blob/master/DOCS/Documentation.md#sd_auto
"""


def sd_command(cmd, *args, **kargs):

    # FILTER KARGS
    if cmd not in KEY_N_ARGS:
        for varg in VALS_N_ARGS:
            if varg in kargs:
                kargs.pop(varg)
    else:
        for varg in VALS_N_ARGS:
            if varg in kargs and varg not in KEY_N_ARGS[cmd]:
                kargs.pop(varg)

    if cmd == 'sd':
        print(SD_COMMANDS_HELP)

    # SD_ENABLE
    elif cmd == 'sd_enable':
        po = kargs.pop('po')
        if po is None:
            print('Pin required, indicate it with -po option')
            see_help(cmd)
        else:
            po = po[0]
            dev = Device(*args, **kargs)
            enabled = dev.cmd(_CMDDICT_['SD_ENABLE'].format(po),
                              silent=True, rtn_resp=True)
            if dev._traceback.decode() in dev.response:
                try:
                    raise DeviceException(dev.response)
                except Exception as e:
                    print(e)
            else:
                if enabled:
                    print('SD Enabled')
                else:
                    print('SD not Enabled')
            dev.disconnect()
        sys.exit()

    # SD_INIT
    elif cmd == 'sd_init':
        dev = Device(*args, **kargs)
        print('Initialzing SD card...')
        sd_mounted = dev.cmd(_CMDDICT_['SD_INIT'],
                             silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if sd_mounted:
                print('SD initiated and mounted Successfully!')
            else:
                print('SD could not be initiated.')
        dev.disconnect()
        sys.exit()

    # SD_DEINIT

    elif cmd == 'sd_deinit':
        dev = Device(*args, **kargs)
        print('Deinitialzing SD card...')
        sd_mounted = dev.cmd(_CMDDICT_['SD_DEINIT'],
                             silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if not sd_mounted:
                print('SD deinitiated Successfully!')
            else:
                print('SD could not be deinitiated.')
        dev.disconnect()
        sys.exit()
    #
    # SD_AUTO

    elif cmd == 'sd_auto':
        dev = Device(*args, **kargs)
        print('Autodetect SD Card mode enabled')
        dev.wr_cmd(_CMDDICT_['SD_AUTO'], follow=True)
        dev.disconnect()
        sys.exit()




from upydev.wsio import wstool, WebSocketFileIO
from upydev.serialio import serialtool, SerialFileIO
from upydev.bleio import bletool, BleFileIO
from upydev.rsyncio import synctool
from upydev.dsyncio import d_sync_recursive, check_wdlog, dev2host_sync_recursive
from upydev.helpinfo import see_help
from upydevice import check_device_type, Device
import upydev
import os
import sys

FILEIO_HELP = """
> FILEIO: Usage '$ upydev ACTION [opts]'
    ACTIONS:
        - put: to upload a file to upy device (see -f, -s, -fre, -dir, -rst, -wdl, -gf)
                e.g. $ upydev put myfile.py, $ upydev put cwd, $ upydev put test_*.py

        - get: to download a file from upy device (see -f, -s, -fre, -dir, -gf)

        - fget: for a faster transfer of large files
            (this needs sync_tool.py in upy device) (see -f, -s, -lh, -wdl, -gf)

        - dsync: to recursively sync a folder in upydevice filesystem
                 where second arg is the directory (can be current working directory too '.',
                 Otherwise use -dir to indicate the folder (must be in cwd).
                 To sync to an Sd card mounted as 'sd' use -s sd.
                 Use -rf to remove files or directories deleted in local dir.
                 Use -d flag to sync from device to host.
                 Use -wdl flag to sync only modified files.

        - rsync: same as "dsync [DIR] -rf -wdl". To recursively sync only modified files.
                 (deleting files too)

        - backup: same as "dsync . -d" to make a backup of the device filesystem.

        - install : install libs to '/lib' path with upip; indicate lib with -f option

        - update_upyutils: to update the latest versions of sync_tool.py, upylog.py,
                        upynotify.py, upysecrets.py, upysh2.py, ssl_repl.py, uping.py, time_it.py,
                        wss_repl.py and wss_helper.py.

"""


def install_w_upip(args, dt, dev_name):
    if args.f is None:
        see_help(args.m)
        sys.exit()
    else:
        dev = Device(args.t, args.p, init=True, ssl=args.wss,
                     auth=args.wss)
        if dt == 'WebSocketDevice':
            print('Installing {} in {} ...'.format(args.f, dev_name))
            dev.wr_cmd("import upip;upip.install('{}');True".format(args.f),
                       silent=True, long_string=True)
            if 'Error' not in dev.response:
                print(dev.response.replace('(\x02ng', 'Installing').replace('True\n', ''),
                      end='')
                result = dev.cmd('_', silent=True, rtn_resp=True)
                if result:
                    print('Done!')
            else:
                print(dev.response.replace('True\n', ''), end='')
            dev.disconnect()
            return

        elif dt == 'SerialDevice':
            sdevIO = SerialFileIO(dev)
            sdevIO.upip_install(args, dev_name)

        elif dt == 'BleDevice':
            bledevIO = BleFileIO(dev)
            bledevIO.upip_install(args, dev_name)

        dev.disconnect()
    return


def update_upyutils(args, dt, dev_name):
    print('Updating device upyutils...')
    utils_dir = os.path.join(upydev.__path__[0], 'upyutils_dir')
    args.m = 'put'
    args.fre = [os.path.join(utils_dir, util) for util in os.listdir(utils_dir)
                if util.endswith('.py')]
    args.dir = 'lib'
    if dt == 'WebSocketDevice':
        wstool(args, dev_name)

    elif dt == 'SerialDevice':
        serialtool(args, dev_name)

    elif dt == 'BleDevice':
        bletool(args, dev_name)


#############################################
def fileio_action(args, **kargs):
    dev_name = kargs.get('device')
    dt = check_device_type(args.t)
    if args.m == 'put' or args.m == 'get':
        if args.wdl and args.m == 'put':
            modified_files, deleted_files = check_wdlog(save_wdlog=True)
            args.fre = modified_files
            if not args.fre:
                # sys.exit()
                return
        if dt == 'WebSocketDevice':
            wstool(args, dev_name)

        elif dt == 'SerialDevice':
            serialtool(args, dev_name)

        elif dt == 'BleDevice':
            bletool(args, dev_name)
    elif args.m == 'update_upyutils':
        update_upyutils(args, dt, dev_name)

    elif args.m == 'install':
        install_w_upip(args, dt, dev_name)

    elif args.m == 'fget':
        if dt == 'WebSocketDevice':
            synctool(args, dev_name)
        else:
            print('Use "get" instead')
    elif args.m == 'dsync' or args.m == 'backup' or args.m == 'rsync':
        if args.m == 'backup':
            args.d = True
            args.f = '.'
        if args.m == 'rsync':
            args.rf = True
            args.wdl = True
        dir_lib = args.f
        dev_lib = args.dir
        dev = Device(args.t, args.p, init=True, ssl=args.wss,
                     auth=args.wss)
        dev.wr_cmd('import os')
        if not dev_lib:
            dev_lib = "./"
        dev_to_host = args.d
        if dt == 'WebSocketDevice':
            wsdevIO = WebSocketFileIO(dev, args, devname=dev_name)
            # print('Syncing @ {} '.format(dev_name), end='\n\n')
            if not dev_to_host:
                print('Syncing to @ {} '.format(dev_name), end='\n\n')
                d_sync_recursive(dir_lib, devIO=wsdevIO,
                                 show_tree=True, rootdir=dev_lib,
                                 root_sync_folder=dir_lib,
                                 args=args,
                                 dev_name=dev_name)
            else:
                print('Syncing from @ {} '.format(dev_name), end='\n\n')
                dev2host_sync_recursive(dir_lib, devIO=wsdevIO,
                                        show_tree=True, rootdir=dev_lib,
                                        root_sync_folder=dir_lib,
                                        args=args,
                                        dev_name=dev_name)
        elif dt == 'SerialDevice':
            sdevIO = SerialFileIO(dev)
            # print('Syncing @ {} '.format(dev_name), end='\n\n')
            if not dev_to_host:
                print('Syncing to @ {} '.format(dev_name), end='\n\n')
                d_sync_recursive(dir_lib, devIO=sdevIO,
                                 show_tree=True,
                                 rootdir=dev_lib,
                                 root_sync_folder=dir_lib,
                                 args=args,
                                 dev_name=dev_name)
            else:
                print('Syncing from @ {} '.format(dev_name), end='\n\n')
                dev2host_sync_recursive(dir_lib, devIO=sdevIO,
                                        show_tree=True,
                                        rootdir=dev_lib,
                                        root_sync_folder=dir_lib,
                                        args=args,
                                        dev_name=dev_name)

        elif dt == 'BleDevice':
            bledevIO = BleFileIO(dev)
            # print('Syncing @ {} '.format(dev_name), end='\n\n')
            if not dev_to_host:
                print('Syncing to @ {} '.format(dev_name), end='\n\n')
                d_sync_recursive(dir_lib, devIO=bledevIO,
                                 show_tree=True,
                                 rootdir=dev_lib,
                                 root_sync_folder=dir_lib,
                                 args=args,
                                 dev_name=dev_name)
            else:
                print('Syncing from @ {} '.format(dev_name), end='\n\n')
                dev2host_sync_recursive(dir_lib, devIO=bledevIO,
                                        show_tree=True,
                                        rootdir=dev_lib,
                                        root_sync_folder=dir_lib,
                                        args=args,
                                        dev_name=dev_name)



import pytest


def pytest_addoption(parser):
    parser.addoption("--dev", action="store", default="default",
                     help="indicate the device with which to run the test")
    parser.addoption("--wss", action="store_true", default=False,
                     help="to indicate use of ssl if WebSecureREPL enabled in WebSocketDevice")


@pytest.fixture
def devname(request):
    return request.config.getoption("--dev")


@pytest.fixture
def use_ssl(request):
    return request.config.getoption("--wss")



from upydevice import Device, DeviceException
from upydev.commandlib import _CMDDICT_
import sys

KEY_N_ARGS = {'wsta_config': ['wp'],
              'wap_config': ['ap']}

VALS_N_ARGS = ['wp', 'ap']

WLAN_UTILS_COMMANDS_HELP = """
> WIFI UTILS: Usage: '$ upydev COMMAND [opts]'
    * (needs wifiutils.py in upydevice, see upyutils in upydev github repo)
    * COMMAND:
        - wlan_init: Initiates wlan util (call this before the following commands)
        - wsta_config: Saves a "netowrk STA" configuration json file in upydevice, use with -wp option as -wp 'ssid' 'password'
        - wap_config: Saves a "netowrk AP" configuration json file in upydevice, use with -ap option as -ap 'ssid' 'password'
        - wsta_conn: Connects to the wlan configured with the command wsta_config
        - wap_conn: Enables the upydevice AP configured with the command wap_config"""


def wlan_command(cmd, *args, **kargs):
    # WLAN UTILS COMMANDS

    # FILTER KARGS
    if cmd not in KEY_N_ARGS:
        for varg in VALS_N_ARGS:
            if varg in kargs:
                kargs.pop(varg)
    else:
        for varg in VALS_N_ARGS:
            if varg in kargs and varg not in KEY_N_ARGS[cmd]:
                kargs.pop(varg)

    # WLAN_INIT
    if cmd == 'wlan_init':
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['WLAN_INIT'])
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            pass
        dev.disconnect()
        sys.exit()

    # WSTA_CONFIG

    elif cmd == 'wsta_config':
        ssid, passwd = kargs.pop('wp')
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['WLAN_CONFIG'].format(ssid, passwd))
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('Wlan {} configured'.format(ssid))
        dev.disconnect()
        sys.exit()

    # WAP_CONFIG

    elif cmd == 'wap_config':
        ssid, passwd = kargs.pop('ap')
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['WLAN_AP_CONFIG'].format(ssid, passwd))
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('access Point {} configured'.format(ssid))
        dev.disconnect()
        sys.exit()

    # WSTA_CONN

    elif cmd == 'wsta_conn':
        dev = Device(*args, **kargs)
        dev.wr_cmd(_CMDDICT_['WLAN_CONN'], follow=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            pass
        dev.disconnect()
        sys.exit()

    # WAP_CONN

    elif cmd == 'wap_conn':
        dev = Device(*args, **kargs)
        dev.wr_cmd(_CMDDICT_['WLAN_AP_CONN'], follow=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            pass
        dev.disconnect()
        sys.exit()

    elif cmd == 'wu':
        print(WLAN_UTILS_COMMANDS_HELP)
        sys.exit()



from upydevice import Device, DeviceException
from upydev.commandlib import _CMDDICT_
import sys
import ast
from datetime import datetime, date
from binascii import hexlify
import time

AUTHMODE_DICT = {0: 'NONE', 1: 'WEP', 2: 'WPA PSK', 3: 'WPA2 PSK',
                    4: 'WPA/WAP2 PSK'}

KEY_N_ARGS = {'du': ['f', 's'], 'df': ['s'], 'netstat_conn': ['wp'],
              'apconfig': ['ap'], 'i2c_config': ['i2c'],
              'spi_config': ['spi'], 'set_ntptime': ['utc'], 'tree': ['f']}

VALS_N_ARGS = ['f', 's', 'wp', 'ap', 'i2c', 'spi', 'utc']


GENERAL_COMMANDS_HELP = """
> GENERAL COMMANDS: Usage '$ upydev COMMAND [opts]'
    COMMAND:
        - info : for upy device system info
        - id : for upy device unique id
        - upysh : to enable the upy shell in the upy device (then do 'upydev man' to
                access upysh manual info)
        - reset : to do a soft reset in upy device
        - kbi : sends CTRL-C signal to stop an ongoing loop, to be able to access repl again
        - uhelp : just calls micropython help
        - umodules: just calls micropython help('modules')
        - meminfo : for upy device RAM memory info; call it once to check actual memory,
                    call it twice and it will free some memory
        - du : to get the size of file in root dir (default) or sd with '-s sd' option;
                    if no file name indicated with -f option, prints all files
        - df : to get memory info of the file system, (total capacity, free, used),
                    (default root dir, -s option to change)
        - tree: to get directory structure in tree format (requires upysh2.py)
        - netinfo : for upy device network info if station is enabled and connected to an AP
        - netinfot : same as netinfo but in table format
        - netscan : for upy device network scan
        - netstat_on : to enable STA
        - netstat_off : to disable STA
        - netstat_conn : to connect to an AP , must provide essid and password (see -wp)
        - netstat : STA state ; returns True if enabled, False if disabled
        - ap_on : to enable AP
        - ap_off : to disable AP
        - apstat : AP state ; returns True if enabled, False if disabled
        - apconfig : AP configuration of essid and password with authmode WPA/WPA2/PSK,
                    (see -ap), needs first that the AP is enabled (do 'upydev ap_on')
        - apscan : scan devices connected to AP; returns number of devices and mac address
        - i2c_config : to configure the i2c pins (see -i2c, defaults are SCL=22, SDA=23)
        - i2c_scan : to scan i2c devices (must config i2c pins first)
        - spi_config: to configure the spi pins (see -spi, defaults are SCK=5,MISO=19,MOSI=18,CS=21)
        - set_localtime: to pass host localtime and set upy device rtc
        - set_ntptime: to set rtc from server, (see -utc for time zone)
        - get_datetime: to get date and time (must be set first, see above commands)"""


def print_sizefile(file_name, filesize, tabs=0):
    _kB = 1024
    if filesize < _kB:
        sizestr = str(filesize) + " by"
    elif filesize < _kB**2:
        sizestr = "%0.1f KB" % (filesize / _kB)
    elif filesize < _kB**3:
        sizestr = "%0.1f MB" % (filesize / _kB**2)
    else:
        sizestr = "%0.1f GB" % (filesize / _kB**3)

    prettyprintname = ""
    for _ in range(tabs):
        prettyprintname += "   "
    prettyprintname += file_name
    print('{0:<40} Size: {1:>10}'.format(prettyprintname, sizestr))


def print_sizefile_all(fileslist, tabs=0, frep=None):
    for filedata in fileslist:
        namefile = filedata[0]
        filesize = filedata[1]

        _kB = 1024
        if filesize < _kB:
            sizestr = str(filesize) + " by"
        elif filesize < _kB**2:
            sizestr = "%0.1f KB" % (filesize / _kB)
        elif filesize < _kB**3:
            sizestr = "%0.1f MB" % (filesize / _kB**2)
        else:
            sizestr = "%0.1f GB" % (filesize / _kB**3)

        prettyprintname = ""
        for _ in range(tabs):
            prettyprintname += "   "
        prettyprintname += namefile
        print('{0:<40} Size: {1:>10}'.format(prettyprintname, sizestr))
        if frep is not None:
            frep.append('{0:<40} Size: {1:>10}'.format(
                prettyprintname, sizestr))


def print_filesys_info(filesize):
    _kB = 1024
    if filesize < _kB:
        sizestr = str(filesize) + " by"
    elif filesize < _kB**2:
        sizestr = "%0.1f KB" % (filesize / _kB)
    elif filesize < _kB**3:
        sizestr = "%0.1f MB" % (filesize / _kB**2)
    else:
        sizestr = "%0.1f GB" % (filesize / _kB**3)
    return sizestr


def _dt_format(number):
        rtc_n = str(number)
        if len(rtc_n) == 1:
            rtc_n = "0{}".format(rtc_n)
            return rtc_n
        else:
            return rtc_n


def _ft_datetime(t_now):
    return([_dt_format(i) for i in t_now])


def gen_command(cmd, *args, **kargs):

    # FILTER KARGS
    if cmd not in KEY_N_ARGS:
        for varg in VALS_N_ARGS:
            if varg in kargs:
                kargs.pop(varg)
    else:
        for varg in VALS_N_ARGS:
            if varg in kargs and varg not in KEY_N_ARGS[cmd]:
                kargs.pop(varg)
    # INFO :
    if cmd == 'info':
        try:
            dev = Device(*args, **kargs)
            print(dev)
            dev.disconnect()
        except Exception as e:
            print(e)
            pass
        return
    # ID:
    elif cmd == 'id':
        dev = Device(*args, **kargs)
        uid = dev.cmd(_CMDDICT_['UID'], silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('ID: {}'.format(uid.decode()))
        dev.disconnect()
        return
    # UPYSH
    elif cmd == 'upysh':
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['UPYSH'], long_string=True)
        dev.disconnect()
        return
    # RESET
    elif cmd == 'reset':
        dev = Device(*args, **kargs)
        dev.reset(reconnect=False)
        time.sleep(0.5)
        dev.disconnect()
        return

    # KEYBOARD Interrupt

    elif cmd == 'kbi':
        dev = Device(*args, **kargs)
        dev.kbi()
        dev.disconnect()
        return

    # UHELP

    elif cmd == 'uhelp':
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['HELP'], long_string=True)
        dev.disconnect()
        return

    # UMODULES

    elif cmd == 'umodules':
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['MOD'], long_string=True)
        dev.disconnect()
        return

    # MEM_INFO

    elif cmd == 'meminfo':
        dev = Device(*args, **kargs)
        RAM = dev.cmd(_CMDDICT_['MEM'], silent=True, rtn_resp=True,
                      long_string=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            mem_info = RAM.splitlines()[1]
            mem = {elem.strip().split(':')[0]: int(elem.strip().split(':')[
                              1]) for elem in mem_info[4:].split(',')}
            print("{0:12}{1:^12}{2:^12}{3:^12}{4:^12}".format(*['Memory',
                                                                'Size', 'Used',
                                                                'Avail',
                                                                'Use%']))
            total_mem = mem['total']/1024
            used_mem = mem['used']/1024
            free_mem = mem['free']/1024
            total_mem_s = "{:.3f} KB".format(total_mem)
            used_mem_s = "{:.3f} KB".format(used_mem)
            free_mem_s = "{:.3f} KB".format(free_mem)

            print('{0:12}{1:^12}{2:^12}{3:^12}{4:>8}'.format('RAM', total_mem_s,
                                                             used_mem_s, free_mem_s,
                                                             "{:.1f} %".format((used_mem/total_mem)*100)))
        dev.disconnect()
        return

    # FILESIZE

    elif cmd == 'du':
        file_name = kargs.pop('f')
        source = kargs.pop('s')
        dev = Device(*args, **kargs)
        # check upysh2:
        is_upysh2 = dev.cmd(_CMDDICT_['CHECK_UPYSH2'], silent=True, rtn_resp=True)
        if is_upysh2:
            if file_name:
                cmd_str = "from upysh2 import du; du('{}')".format(file_name)
            else:
                cmd_str = "from upysh2 import du; du"
            dev.wr_cmd(cmd_str, follow=True)
        else:

            if source is not None:
                file_name = '/{}/{}'.format(source, file_name)
            cmd_str = _CMDDICT_['OS_STAT'].format(file_name)
            if file_name is None:
                dir = ''
                if source is not None:
                    dir = '/{}/'.format(source)
                cmd_str = _CMDDICT_['FILE_STAT'].format(dir)
            du_info = dev.cmd(cmd_str, silent=True, rtn_resp=True)
            if dev._traceback.decode() in dev.response:
                try:
                    raise DeviceException(dev.response)
                except Exception as e:
                    print(e)
            else:
                if len(du_info) > 0:
                    if file_name is not None:
                        size = du_info[6]
                        print_sizefile(file_name, size)
                    else:
                        print_sizefile_all(du_info)
        dev.disconnect()
        return

    # FILESYS_INFO

    elif cmd == 'df':
        source = kargs.pop('s')
        dev = Device(*args, **kargs)
        filesys = ''
        filesys_check = True
        if source is not None:
            filesys = '/{}'.format(source)
            filesys_check = dev.cmd(_CMDDICT_['CHECK_DIR'].format(filesys.split('/')[1]),
                                    silent=True,
                                    rtn_resp=True)
            if dev._traceback.decode() in dev.response:
                try:
                    raise DeviceException(dev.response)
                except Exception as e:
                    print(e)
        cmd_str = _CMDDICT_['STAT_FS'].format(filesys)
        if filesys == '':
            filesys = 'Flash'
        if filesys_check:
            resp = dev.cmd(cmd_str, silent=True, rtn_resp=True)
            if dev._traceback.decode() in dev.response:
                try:
                    raise DeviceException(dev.response)
                except Exception as e:
                    print(e)
            else:
                size_info = resp
                total_b = size_info[0] * size_info[2]
                used_b = (size_info[0] * size_info[2]) - (size_info[0] * size_info[3])
                total_mem = print_filesys_info(total_b)
                free_mem = print_filesys_info(size_info[0] * size_info[3])
                used_mem = print_filesys_info(used_b)
                if filesys == 'Flash':
                    mounted_on = '/'
                else:
                    mounted_on = filesys
                print("{0:12}{1:^12}{2:^12}{3:^12}{4:^12}{5:^12}".format(*['Filesystem',
                                                                           'Size', 'Used',
                                                                           'Avail',
                                                                           'Use%', 'Mounted on']))
                print('{0:12}{1:^12}{2:^12}{3:^12}{4:>8}{5:>5}{6:12}'.format(filesys, total_mem,
                                                                             used_mem, free_mem,
                                                                             "{:.1f} %".format((used_b/total_b)*100), ' ', mounted_on))
        else:
            print('{} not mounted'.format(filesys))
        dev.disconnect()
        return

    # TREE
    elif cmd == 'tree':
        dir_name = kargs.pop('f')
        dev = Device(*args, **kargs)
        # check upysh2:
        is_upysh2 = dev.cmd(_CMDDICT_['CHECK_UPYSH2'], silent=True, rtn_resp=True)
        if is_upysh2:
            if dir_name:
                cmd_str = "from upysh2 import tree; tree('{}')".format(dir_name)
            else:
                cmd_str = "from upysh2 import tree; tree"
            dev.wr_cmd(cmd_str, follow=True)
        else:
            pass
        dev.disconnect()
        return

    # NETINFO
    elif cmd == 'netinfo':
        dev = Device(*args, **kargs)
        net_info_list = dev.cmd(_CMDDICT_['NET_INFO'], silent=True,
                                rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if len(net_info_list) > 0:
                print('IF CONFIG: {}'.format(net_info_list[0]))
                print('SSID: {}'.format(net_info_list[1]))
                vals = hexlify(net_info_list[2]).decode()
                bssid = ':'.join([vals[i:i+2] for i in range(0, len(vals), 2)])
                print('BSSID: {}'.format(bssid))
                print('RSSI: {} dBm'.format(net_info_list[3]))
        dev.disconnect()
        return

    # NETINFOT
    elif cmd == 'netinfot':
        dev = Device(*args, **kargs)
        net_info_list = dev.cmd(_CMDDICT_['NET_INFO'], silent=True,
                                rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if len(net_info_list) > 0:
                vals = hexlify(net_info_list[2]).decode()
                bssid = ':'.join([vals[i:i+2] for i in range(0, len(vals), 2)])
                print('IF CONFIG:')
                print('┏{0}━┳━{1}━┳━{2}━┳━{3}━┳━{4}━┳━{5}━┳━{6}━┓'.format(
                                '━'*15, '━'*15, '━'*15, '━'*15, '━'*15, '━'*20,
                                '━'*10))
                print('┃{0:^15} ┃ {1:^15} ┃ {2:^15} ┃ {3:^15} ┃ {4:^15} ┃ {5:^20} ┃ {6:^10} ┃'.format(
                                'IP', 'SUBNET', 'GATEAWAY', 'DNS', 'SSID', 'BSSID',
                                'RSSI'))
                print('┣{0}━╋━{1}━╋━{2}━╋━{3}━╋━{4}━╋━{5}━╋━{6}━┫'.format(
                                '━'*15, '━'*15, '━'*15, '━'*15, '━'*15, '━'*20,
                                '━'*10))
                try:
                    print('┃{0:^15} ┃ {1:^15} ┃ {2:^15} ┃ {3:^15} ┃ {4:^15} ┃ {5:^20} ┃ {6:^10} ┃'.format(
                        *net_info_list[0], net_info_list[1],
                        bssid, net_info_list[3]))
                    print('┗{0}━┻━{1}━┻━{2}━┻━{3}━┻━{4}━┻━{5}━┻━{6}━┛'.format(
                                '━'*15, '━'*15, '━'*15, '━'*15, '━'*15, '━'*20,
                                '━'*10))
                except Exception as e:
                    print(e)
        dev.disconnect()
        return

    # NETSCAN
    elif cmd == 'netscan':
        dev = Device(*args, **kargs)
        if dev.dev_class == 'SerialDevice':
            dev.cmd(_CMDDICT_['NET_STAT_ON'], silent=True)
        net_scan_list = dev.cmd(_CMDDICT_['NET_SCAN'],
                                silent=True,
                                rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            # FIX FOR SERIAL ESP DEBUGGING INFO
            if isinstance(net_scan_list, str):
                try:
                    net_scan_list = ast.literal_eval(net_scan_list.split('[0m')[1])
                except Exception as e:
                    net_scan_list = []
            print('┏{0}━┳━{1}━┳━{2}━┳━{3}━┳━{4}━┳━{5}━┓'.format(
                            '━'*20, '━'*25, '━'*10, '━'*15, '━'*15, '━'*10))
            print('┃{0:^20} ┃ {1:^25} ┃ {2:^10} ┃ {3:^15} ┃ {4:^15} ┃ {5:^10} ┃'.format(
                'SSID', 'BSSID', 'CHANNEL', 'RSSI', 'AUTHMODE', 'HIDDEN'))
            print('┣{0}━╋━{1}━╋━{2}━╋━{3}━╋━{4}━╋━{5}━┫'.format(
                            '━'*20, '━'*25, '━'*10, '━'*15, '━'*15, '━'*10))
            for net in net_scan_list:
                netscan = net
                auth = AUTHMODE_DICT[netscan[4]]
                vals = hexlify(netscan[1]).decode()
                ap_name = netscan[0].decode()
                if len(ap_name) > 20:
                    ap_name = ap_name[:17] + '...'
                bssid = ':'.join([vals[i:i+2] for i in range(0, len(vals), 2)])
                print('┃{0:^20} ┃ {1:^25} ┃ {2:^10} ┃ {3:^15} ┃ {4:^15} ┃ {5:^10} ┃'.format(
                    ap_name, bssid, netscan[2], netscan[3],
                    auth, str(netscan[5])))
            print('┗{0}━┻━{1}━┻━{2}━┻━{3}━┻━{4}━┻━{5}━┛'.format(
                        '━'*20, '━'*25, '━'*10, '━'*15, '━'*15, '━'*10))
        dev.disconnect()
        return

    # NETSTAT_ON

    elif cmd == 'netstat_on':
        dev = Device(*args, **kargs)
        stat_on = dev.cmd(_CMDDICT_['NET_STAT_ON'], silent=True,
                          rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if stat_on:
                print('Station Enabled')
        dev.disconnect()
        return

    # NETSTAT_OFF

    elif cmd == 'netstat_off':
        dev = Device(*args, **kargs)
        stat_off = dev.cmd(_CMDDICT_['NET_STAT_OFF']+'\r'*5, silent=True,
                           rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('Station Disabled')
        dev.disconnect()
        return

    # NETSTAT_CONN

    elif cmd == 'netstat_conn':
        ssid, passwd = kargs.pop('wp')
        print('Connecting to {}...'.format(ssid))
        connect_to = _CMDDICT_['NET_STAT_CONN'].format(ssid, passwd)
        dev = Device(*args, **kargs)
        dev.cmd(connect_to, silent=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            pass
        dev.disconnect()
        return

    # NETSTAT
    elif cmd == 'netstat':
        dev = Device(*args, **kargs)
        netstat = dev.cmd(_CMDDICT_['NET_STAT'], silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if netstat:
                print('Station Enabled')
            else:
                print('Station Disabled')
            pass
        dev.disconnect()
        return

    # AP_ON

    elif cmd == 'ap_on':
        dev = Device(*args, **kargs)
        ap_on = dev.cmd(_CMDDICT_['AP_ON'], silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if ap_on:
                print('access Point Enabled')
            else:
                print('access Point Disabled')
            pass
        dev.disconnect()
        return
    #
    # AP_OFF

    elif cmd == 'ap_off':
        dev = Device(*args, **kargs)
        ap_off = dev.cmd(_CMDDICT_['AP_OFF'], silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if not ap_off:
                print('access Point Disabled')
            else:
                print('access Point Enabled')
            pass
        dev.disconnect()
        return

    # APSTAT

    elif cmd == 'apstat':
        dev = Device(*args, **kargs)
        apstat = dev.cmd(_CMDDICT_['APSTAT'], silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if len(apstat) > 0:
                auth = AUTHMODE_DICT[int(apstat[-2])]
                print('AP INFO:')
                print('┏{0}━┳━{1}━┳━{2}━┳━{3}━┓'.format(
                                '━'*20, '━'*20, '━'*20, '━'*20))
                print('┃{0:^20} ┃ {1:^20} ┃ {2:^20} ┃ {3:^20} ┃'.format(
                    'AP ENABLED', 'ESSID', 'CHANNEL', 'AUTHMODE'))
                print('┃{0:^20} ┃ {1:^20} ┃ {2:^20} ┃ {3:^20} ┃'.format(
                    *apstat[:-2], auth))
                print('┣{0}━╋━{1}━╋━{2}━╋━{3}━┫'.format(
                                '━'*20, '━'*20, '━'*20, '━'*20))
                print('┃{0:^20} ┃ {1:^20} ┃ {2:^20} ┃ {3:^20} ┃'.format(
                    'IP', 'SUBNET', 'GATEAWAY', 'DNS'))
                try:
                    print('┃{0:^20} ┃ {1:^20} ┃ {2:^20} ┃ {3:^20} ┃'.format(
                        *apstat[-1]))
                    print('┗{0}━┻━{1}━┻━{2}━┻━{3}━┛'.format(
                                '━'*20, '━'*20, '━'*20, '━'*20))
                except Exception as e:
                    print(e)
                    pass
        dev.disconnect()
        return

    # APCONFIG

    elif cmd == 'apconfig':
        ssid, passwd = kargs.pop('ap')
        print('Configuring {} access Point ...'.format(ssid))
        apconfig = _CMDDICT_['AP_CONFIG'].format(ssid, passwd)
        if len(passwd) < 8:
            print('[WARNING]: Password too short (less than 8 characters)')
        dev = Device(*args, **kargs)
        dev.cmd(apconfig, silent=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('{} access Point Configured'.format(ssid))
        dev.disconnect()
        return

    # APSCAN

    elif cmd == 'apscan':
        dev = Device(*args, **kargs)
        ap_scan_list = dev.cmd(_CMDDICT_['AP_SCAN'], silent=True,
                               rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if len(ap_scan_list) > 0:
                ap_devices = ap_scan_list
                print('Found {} devices:'.format(len(ap_devices)))
                for cdev in ap_devices:
                    bytdev = hexlify(cdev[0]).decode()
                    mac_ad = ':'.join([bytdev[i:i+2] for i in range(0, len(bytdev),
                                                                    2)])
                    print('MAC: {}'.format(mac_ad))
            else:
                print('No device found')
        dev.disconnect()
        return

    # I2C_CONFIG # FIX FOR PYBOARD, AUTODETECT

    elif cmd == 'i2c_config':
        scl, sda = kargs.pop('i2c')
        dev = Device(*args, **kargs, autodetect=True)
        if dev.dev_platform == 'pyboard':
            dev.cmd(_CMDDICT_['I2C_CONFIG_PYB'].format(scl, sda), silent=True)
        else:
            dev.cmd(_CMDDICT_['I2C_CONFIG'].format(scl, sda), silent=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('I2C configured:\nSCL = Pin({}), SDA = Pin({})'.format(scl,
                                                                         sda))
        dev.disconnect()
        return

    # I2C_SCAN

    elif cmd == 'i2c_scan':
        dev = Device(*args, **kargs)
        i2c_scan_list = dev.cmd(_CMDDICT_['I2C_SCAN'], silent=True,
                                rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            if len(i2c_scan_list) > 0:
                i2c_devices = i2c_scan_list
                print('Found {} devices:'.format(len(i2c_devices)))
                print(i2c_devices)
                print('Hex:')
                print([hex(i2c_dev) for i2c_dev in i2c_devices])
            else:
                print('No device found')
        dev.disconnect()
        return
    #
    # SPI CONFIG
    elif cmd == 'spi_config':
        spi_conf = kargs.pop('spi')
        dev = Device(*args, **kargs)
        dev.cmd(_CMDDICT_['SPI_CONFIG'].format(*spi_conf), silent=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('SPI configured:\nSCK = Pin({}), MISO = Pin({}), MOSI = Pin({}), CS = Pin({})'.format(
                    *spi_conf))
        dev.disconnect()
        return

    #  * RTC *

    # SET LOCAL TIME

    elif cmd == 'set_localtime':
        dev = Device(*args, **kargs)
        print('Setting local time: {}'.format(
            datetime.now().strftime("%Y-%m-%d T %H:%M:%S")))
        wkoy = date.today().isocalendar()[1]
        datetime_local = [val for val in datetime.now().timetuple()[:-3]]
        datetime_tuple = datetime_local[:3]
        datetime_tuple.append(wkoy)
        datetime_final = datetime_tuple + datetime_local[3:] + [0]
        dev.cmd(_CMDDICT_['SET_RTC_LT'].format(*datetime_final), silent=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            print('Done!')
        dev.disconnect()
        return
    #
    # SET NTP TIME
    elif cmd == 'set_ntptime':
        utc = kargs.pop('utc')
        print('Setting time UTC+{} from NTP Server'.format(utc))
        dev = Device(*args, **kargs)
        for ntp_cmd in _CMDDICT_['SET_RTC_NT'].format(utc).split(';'):
            dev.cmd(ntp_cmd, silent=True)
            if dev._traceback.decode() in dev.response:
                try:
                    raise DeviceException(dev.response)
                except Exception as e:
                    print(e)

        print('Done!')
        dev.disconnect()
        return

    # GET UPY DEVICE LOCALTIME
    elif cmd == 'get_datetime':
        dev = Device(*args, **kargs)
        datetime_dev = dev.cmd(_CMDDICT_['DATETIME'],
                               silent=True, rtn_resp=True)
        if dev._traceback.decode() in dev.response:
            try:
                raise DeviceException(dev.response)
            except Exception as e:
                print(e)
        else:
            formatted_devtime = _ft_datetime(datetime_dev)
            print('Device RTC time: {}-{}-{} T {}:{}:{}'.format(*formatted_devtime))
        dev.disconnect()
        return

    elif cmd == 'gc':
        print(GENERAL_COMMANDS_HELP)
        sys.exit()



from upydevice import Device
import sys
from upydev.helpinfo import see_help
from upydevice import check_device_type
import getpass
import os
import json
import upydev
from binascii import hexlify
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives import serialization
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
import string
import secrets
import hashlib
import netifaces
import socket
import locale
from datetime import datetime, timedelta
import shlex
import subprocess
import ipaddress


UPYDEV_PATH = upydev.__path__[0]

KEYGEN_HELP = """
> KEYGEN: Usage '$ upydev ACTION [opts]'
    ACTIONS:
        - gen_rsakey: To generate RSA-2048 bit key that will be shared with the device
                      (it is unique for each device) use -tfkey to send this key to the
                      device (use only if connected directly to the AP of the device or a
                      "secure" wifi e.g. local/home). If not connected to a "secure" wifi
                      upload the key (it is stored in upydev.__path__) by USB/Serial connection.

        - rf_wrkey: To "refresh" the WebREPL password with a new random password derivated from
                    the RSA key previously generated. A token then is sent to the device to generate
                    the same password from the RSA key previously uploaded. This won't leave
                    any clues in the TCP Websocekts packages of the current WebREPL password.
                    (Only the token will be visible; check this using wireshark)
                    (This needs upysecrets.py)

        - sslgen_key: (This needs openssl available in $PATH)
                       To generate ECDSA key and a self-signed certificate to enable SSL sockets
                       This needs a passphrase, that will be required every time the key is loaded.
                       Use -tfkey to upload this key to the device
                       (use only if connected directly to the AP of the device or a
                       "secure" wifi e.g. local/home). If not connected to a "secure" wifi
                       upload the key (it is stored in upydev.__path__) by USB/Serial connection.
                       Use -to [serial devname] flag with -tfkey to transfer keys by USB/Serial
"""


# CRYPTO


def rsa_keygen(args, dir='', store=True, show_key=False, id='00'):
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=args.key_size,
        backend=default_backend())
    my_p = getpass.getpass(prompt='Password: ', stream=None)
    pem = private_key.private_bytes(encoding=serialization.Encoding.PEM, format=serialization.PrivateFormat.PKCS8,
                                    encryption_algorithm=serialization.BestAvailableEncryption(bytes(my_p, 'utf-8')))
    if show_key:
        print(pem)
    if store:
        key_path_file = os.path.join(dir, 'upy_pub_rsa{}.key'.format(id))
        with open(key_path_file, 'wb') as keyfile:
            keyfile.write(pem)
    return pem


def load_rsa_key(dir='', show_key=False, id='00'):
    buff_key = b''
    rsa_key_abs = os.path.join(dir, 'upy_pub_rsa{}.key'.format(id))
    try:
        with open(rsa_key_abs, 'rb') as keyfile:
            while True:
                try:
                    buff = keyfile.read(2000)
                    if buff != b'':
                        buff_key += buff
                    else:
                        break
                except Exception as e:
                    print(e)
        if show_key:
            print(buff_key)
        return buff_key
    except Exception as e:
        print("No RSA key found for this device, generate one first with '$ upydev gen_rsakey -tfkey' ")


def upy_keygen(rsa_key):
    aZ09 = bytes(string.ascii_letters + string.digits, 'ascii')
    raw_key_list = [line for line in rsa_key.splitlines()[1:-1]]
    raw_key = b''
    for line in raw_key_list:
        raw_key += line
    random_token = secrets.token_bytes(32)  # send this
    for b in random_token:
        raw_key += bytes(chr(raw_key[b]), 'utf-8')
    key_hash = hashlib.sha256()
    key_hash.update(raw_key)
    hashed_key = key_hash.digest()
    index_key = [secrets.randbelow(len(hashed_key)) for i in range(8)]  # send this
    password_long = bytes([hashed_key[val] for val in index_key])
    password_short = bytes([aZ09[val % len(aZ09)] for val in password_long]).decode()

    return (password_short, random_token + bytes(index_key))


def get_cert_data():
    # MAC ADDRES IF
    try:
        addrs = [netifaces.ifaddresses(iface)[netifaces.AF_INET][0]['addr'] for
                 iface in netifaces.interfaces() if netifaces.AF_INET in
                 netifaces.ifaddresses(iface)][-1]

        USER = os.environ['USER']
        HOST_NAME = socket.gethostname()
        COUNTRY_CODE = locale.getlocale()[0].split('_')[1]
        return {'addrs': addrs, 'USER': USER, 'HOST_NAME': HOST_NAME,
                'COUNTRY_CODE': COUNTRY_CODE}
    except Exception as e:
        default_temp = {'addrs': 'xxxx', 'USER': 'xxxx', 'HOST_NAME': 'xxxx',
                        'COUNTRY_CODE': 'XX'}
        return default_temp


def ssl_ECDSA_key_certgen(args, dir='', store=True):
    # print('Generating RSA key and Cerficate...')
    print('Getting unique id...')
    dev = Device(args.t, args.p, init=True, ssl=args.wss, auth=args.wss,
                 autodetect=True)
    id_bytes = dev.cmd("from machine import unique_id; unique_id()",
                       silent=True, rtn_resp=True)
    unique_id = hexlify(id_bytes).decode()
    print('ID: {}'.format(unique_id))
    dev_platform = dev.dev_platform
    dev.disconnect()
    # key = rsa.generate_private_key(
    #     public_exponent=65537,
    #     key_size=args.key_size,
    #     backend=default_backend())
    key = ec.generate_private_key(
        ec.SECP256R1(),
        backend=default_backend())

    my_p = getpass.getpass(prompt='Passphrase: ', stream=None)
    pem = key.private_bytes(encoding=serialization.Encoding.PEM,
                            format=serialization.PrivateFormat.TraditionalOpenSSL,
                            encryption_algorithm=serialization.BestAvailableEncryption(bytes(my_p, 'utf-8')))

    if store:
        key_path_file_pem = os.path.join(dir, 'SSL_key{}.pem'.format(unique_id))
        with open(key_path_file_pem, 'wb') as keyfile:
            keyfile.write(pem)
    cert_data = get_cert_data()
    subject = issuer = x509.Name([x509.NameAttribute(NameOID.COUNTRY_NAME, u"{}".format(cert_data['COUNTRY_CODE'])),
                                  x509.NameAttribute(
                                      NameOID.USER_ID, u"{}".format(cert_data['USER'])),
                                  x509.NameAttribute(
                                      NameOID.SURNAME, u"{}".format(cert_data['HOST_NAME'])),
                                  x509.NameAttribute(
                                      NameOID.STREET_ADDRESS, u"{}".format(cert_data['addrs'])),
                                  x509.NameAttribute(
                                      NameOID.ORGANIZATION_NAME, u"MicroPython"),
                                  x509.NameAttribute(NameOID.COMMON_NAME, u"{}@{}".format(dev_platform, unique_id))])
    host_ip = ipaddress.IPv4Address(cert_data['addrs'])
    cert = x509.CertificateBuilder().subject_name(
                subject).issuer_name(issuer).public_key(key.public_key()
                                                        ).serial_number(x509.random_serial_number()
                                                                        ).not_valid_before(datetime.utcnow()
                                                                                           ).not_valid_after(datetime.utcnow() + timedelta(days=365)
                                                                                                             ).add_extension(x509.SubjectAlternativeName([x509.DNSName(u"localhost"),
                                                                                                                                                          x509.IPAddress(host_ip),
                                                                                                                                                          x509.DNSName(
                                                                                                                 u"wss://{}:8833".format(args.t)),
                                                                                                                 x509.DNSName(u"wss://192.168.4.1:8833")]),
                                                  critical=False
                                                  ).sign(key, hashes.SHA256(), default_backend())
    if store:
        cert_path_file_pem = os.path.join(dir,
                                          'SSL_certificate{}.pem'.format(unique_id))
        with open(cert_path_file_pem, 'wb') as certfile:
            certfile.write(cert.public_bytes(serialization.Encoding.PEM))

    # CONVERT TO DER FORMAT
    cert_path_file_der = os.path.join(dir, 'SSL_certificate{}.der'.format(unique_id))
    cert_args = shlex.split(
        'openssl x509 -in {} -out {} -outform DER'.format(cert_path_file_pem,
                                                          cert_path_file_der))
    subprocess.call(cert_args)
    key_path_file_der = os.path.join(dir, 'SSL_key{}.der'.format(unique_id))
    key_args = shlex.split(
        'openssl ec -in {} -out {} -outform DER'.format(key_path_file_pem,
                                                        key_path_file_der))
    subprocess.call(key_args)

    if args.tfkey:
        print('Transfering ECDSA key and Certificates to the device...')
        ssl_k = key_path_file_der
        ssl_c = cert_path_file_der
        return {'ACTION': 'put', 'mode': 'SSL', 'Files': [ssl_k, ssl_c]}


# CRYPTO


def get_rsa_key(args):
    print('Generating RSA key...')
    print('Getting unique id...')
    dev = Device(args.t, args.p, init=True, ssl=args.wss, auth=args.wss)
    id_bytes = dev.cmd("from machine import unique_id; unique_id()",
                       silent=True, rtn_resp=True)
    dev.disconnect()

    unique_id = hexlify(id_bytes).decode()
    print('ID: {}'.format(unique_id))
    pkey = rsa_keygen(args, dir=UPYDEV_PATH, show_key=args.show_key, id=unique_id)
    print('Done!')

    if args.tfkey:  # move out for now
        print('Transfering RSA key to the device...')
        key_abs = os.path.join(UPYDEV_PATH, 'upy_pub_rsa{}.key'.format(unique_id))
        return {'ACTION': 'put', 'mode': 'RSA', 'Files': [key_abs]}
        # print('Done!')


def refresh_wrkey(args, device):
    print('Generating new random WebREPL password for {}...'.format(device))
    if args.localkey_id is None:
        print('Getting unique id...')
        dev = Device(args.t, args.p, init=True, ssl=args.wss, auth=args.wss)
        id_bytes = dev.cmd("from machine import unique_id; unique_id()",
                           silent=True, rtn_resp=True)
        unique_id = hexlify(id_bytes).decode()
        print('ID: {}'.format(unique_id))
        pvkey = load_rsa_key(dir=UPYDEV_PATH, show_key=args.show_key,
                             id=unique_id)
        print('Generating random password from RSA key...')
        new_p, new_token = upy_keygen(pvkey)
        dev.cmd("from upysecrets import load_key, upy_keygen")
        dev.cmd("rk = load_key()")
        dev.cmd("newp = upy_keygen(rk, {})".format(new_token))
        print('New random password generated!')
        dev.reset(reconnect=False)
    else:
        unique_id = args.localkey_id
        print('ID: {}'.format(unique_id))
        pvkey = load_rsa_key(dir=UPYDEV_PATH, show_key=args.show_key,
                             id=unique_id)
        print('Generating random password from RSA key...')
        new_p, new_token = upy_keygen(pvkey)
        print('Use this token to generate new password in the device:\n', new_token)
    upydev_ip = args.t
    upydev_pass = new_p
    upy_conf = {'addr': upydev_ip, 'passwd': upydev_pass, 'name': device}
    file_conf = 'upydev_.config'
    if args.g:
        file_conf = os.path.join(UPYDEV_PATH, 'upydev_.config')
    with open(file_conf, 'w') as config_file:
        config_file.write(json.dumps(upy_conf))
    if args.g:
        print('{} settings saved globally!'.format(device))
    else:
        print('{} settings saved in working directory!'.format(device))
    # check if device is in global group
    dt = check_device_type(args.t)
    upydev_name = device
    group_file = 'UPY_G.config'
    group_file_path = os.path.join(UPYDEV_PATH, group_file)
    if group_file in os.listdir(UPYDEV_PATH):
        with open(group_file_path, 'r', encoding='utf-8') as group:
            devices = json.loads(group.read())

        if upydev_name in devices:
            devices.update({upydev_name: [upydev_ip, upydev_pass]})
            with open(group_file_path, 'w', encoding='utf-8') as group:
                group.write(json.dumps(devices))
            print('{} {} settings saved in global group!'.format(dt, upydev_name))


def get_ssl_keycert(args):
    # print('Generating SSL ECDSA key and certificates...')
    ssl_dict = ssl_ECDSA_key_certgen(args, dir=UPYDEV_PATH)
    if ssl_dict:
        return ssl_dict


def keygen_action(args, **kargs):
    dev_name = kargs.get('device')
    # GEN_RSAKEY

    if args.m == 'gen_rsakey':
        print('Generating RSA key for {}...'.format(dev_name))
        rsa_key = get_rsa_key(args)
        if rsa_key:
            return rsa_key
        else:
            sys.exit()

    # RF_WRKEY
    elif args.m == 'rf_wrkey':
        refresh_wrkey(args, dev_name)
        sys.exit()

    # SSLGEN_KEY

    elif args.m == 'sslgen_key':
        print('Generating SSL ECDSA key, cert for : {}'.format(dev_name))
        ssl_key_cert = get_ssl_keycert(args)
        if ssl_key_cert:
            return ssl_key_cert
        else:
            sys.exit()



import upydev
import json
import os
import textwrap
import webbrowser

UPYDEV_PATH = upydev.__path__[0]


HELP_INFO_ARG = '''Mode/Tools:
> DEVICE MANAGEMENT: '$ upydev dm' to see help on device management.
    ACTIONS : config, check, set, make_group, mg_group, see, gg

> FILEIO: '$ upydev fio' to see help on file input/ouput operations.
    ACTIONS: put, get, fget, dsync, rsync, backup, install, update_upyutils

> FIRMWARE: '$ upydev fw' to see help on firmware operations.
    ACTIONS: fwr, flash, mpyx

> KEYGEN: '$ upydev kg' to see help on keygen operations.
    ACTIONS: gen_rsakey, rf_wrkey, sslgen_key

> REPLS: '$ upydev rp' to see help on repls modes.
    ACTIONS: repl, rpl, wrepl, wssrepl, srepl

> SHELL-REPLS: '$ upydev sh' to see help on shell-repls modes.
    ACTIONS: shell, shl, ssl_wrepl, ssl, sh_srepl, shr, wssl, set_wss, ble, jupyterc

> DEBUGGING: '$ upydev db' to see help on debugging operations.
    ACTIONS: ping, probe, scan, run, timeit, diagnose, errlog, stream_test,
             sysctl, log, debug, pytest-setup, pytest

> GROUP COMMAND MODE: '$ upydev gp' to see help on group mode options.
    OPTIONS: -G, -GP

> HELP: '$ upydev h' or '$ upydev help' to see help (without optional args)
        '$ upydev -h' or '$ upydev --help' to see full help info.

        - To see help about a any ACTION/COMMAND
          put %% before that ACTION/COMMAND as : $ upydev %%ACTION

    ACTIONS: help, h, dm, fio, fw, kg, rp, sh, db, gp, gc, wu, sd, pro, docs,
             udocs, mdocs.

upy Commands:
> GENERAL: do '$ upydev gc' to see General commmands help.

> WIFI UTILS: do '$ upydev wu' to see Wifi utils commands help.

> SD: do '$ upydev sd' to see SD utils commands help.

> PROTOTYPE: do '$ upydev pro' to see Prototype utils commands help.
'''


def see_help(cmd):
    help_file = os.path.join(UPYDEV_PATH, 'help.config')
    with open(help_file, 'r') as helpref:
        help_dict = json.loads(helpref.read())
    columns, rows = os.get_terminal_size(0)
    if cmd is not None:
        if cmd in help_dict:
            print('\n'.join(textwrap.wrap(help_dict[cmd], columns-3)))
        else:
            print('Help info not available for "{}" command'.format(cmd))
    else:
        pass


def see_docs(args):
    if args.m == 'mdocs':
        docs_url = "docs.micropython.org"
    elif args.m == 'docs':
        docs_url = "upydev.readthedocs.io"
    elif args.m == 'udocs':
        docs_url = "upydevice.readthedocs.io"
    if args.f:
        key_word = args.f
        search = f'https://{docs_url}/en/latest/search.html?q={key_word}&check_keywords=yes&area=default'
        webbrowser.open(search)
    else:
        webbrowser.open(f'https://{docs_url}/en/latest/')



#!/usr/bin/env python
# Do sync , d_sync and backup
# @Author: carlosgilgonzalez
# @Date:   2019-07-09T20:49:15+01:00
# @Last modified by:   carlosgilgonzalez
# @Last modified time: 2019-07-13T22:18:43+01:00

import socket
import time as time_h
import os
import netifaces
import sys
from datetime import timedelta
from upydevice import Device, DeviceNotFound, DeviceException
from upydev.helpinfo import see_help


class SyncFileIO:
    def __init__(self, dev, port=None):
        self.host = None
        self.port = port
        self.dev = dev
        self.buff = bytearray(1024*2)
        self.bloc_progress = ["▏", "▎", "▍", "▌", "▋", "▊", "▉"]
        self.columns, self.rows = os.get_terminal_size(0)
        self.cnt_size = 65
        self.bar_size = int((self.columns - self.cnt_size))
        self.pb = False
        self.wheel = ['|', '/', '-', "\\"]
        self.server_ip = self.get_ip()
        self.server_socket = None
        self.conn = None

    def get_pb(self):
        self.columns, self.rows = os.get_terminal_size(0)
        if self.columns > self.cnt_size:
            self.bar_size = int((self.columns - self.cnt_size))
            self.pb = True
        else:
            self.bar_size = 1
            self.pb = False

    def do_pg_bar(self, index, wheel, nb_of_total, speed, time_e, loop_l,
                  percentage, ett):
        l_bloc = self.bloc_progress[loop_l]
        if index == self.bar_size:
            l_bloc = "█"
        sys.stdout.write("\033[K")
        print('▏{}▏{:>2}{:>5} % | {} | {:>5} KB/s | {}/{} s'.format("█" *index + l_bloc  + " "*((self.bar_size+1) - len("█" *index + l_bloc)),
                                                                        wheel[index % 4],
                                                                        int((percentage)*100),
                                                                        nb_of_total, speed,
                                                                        str(timedelta(seconds=time_e)).split('.')[0][2:],
                                                                        str(timedelta(seconds=ett)).split('.')[0][2:]), end='\r')
        sys.stdout.flush()

    def get_ip(self):
        try:
            ip = [netifaces.ifaddresses(iface)[netifaces.AF_INET][0]['addr'] for
                  iface in netifaces.interfaces() if netifaces.AF_INET in
                  netifaces.ifaddresses(iface)][-1]
            return ip
        except Exception as e:
            try:
                ip_soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                ip_soc.connect(('8.8.8.8', 1))
                local_ip = ip_soc.getsockname()[0]
                ip_soc.close()
                return local_ip
            except Exception as e:
                return '0.0.0.0'

    def connect(self, args):

        self.dev.wr_cmd('import sync_tool as synct', silent=True, follow=False)

        time_h.sleep(0.2)
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((self.server_ip, self.port))
        self.server_socket.listen(1)
        # print('Server listening...')
        sync_connect_cmd = "cli_soc = synct.connect_SOC('{}', {})".format(self.server_ip,
                                                                          self.port)
        self.dev.wr_cmd(sync_connect_cmd, silent=True, follow=False)
        self.conn, addr = self.server_socket.accept()
        # print('Connection received from {}...'.format(addr))
        self.conn.settimeout(5)

    def sync_file(self, args, dev_name, out=None):
        self.get_pb()
        dir = ''
        if args.dir:
            dir = args.dir
        filetoget = args.f
        cmd_filesize_str = "synct.os.stat('{}' + '/' + '{}')[6]".format(
            dir, filetoget)
        # espdev.output = None
        # while espdev.output is None:
        filesize = self.dev.wr_cmd(cmd_filesize_str, silent=True, follow=False,
                                   rtn_resp=True)
        print('File size: {:>40.2f} kB'.format(filesize/1024))
        time_h.sleep(1)
        sync_tool_cmd = "synct.read_sd_file_sync_raw({},cli_soc)".format(
            "'{}/{}'".format(dir, filetoget))
        self.dev.cmd_nb(sync_tool_cmd)
        print('Syncing file {} from {}'.format(filetoget, dev_name))
        print('Saving {} file'.format(filetoget))
        print('Progress:\n')
        t0 = 0
        buff = b''
        t_start = time_h.time()
        EOF = False
        n_EOF = 1
        EOF_count = 0
        if out is not None:
            filetoget = out
        with open(filetoget, 'wb') as log:
            pass
        while True:
            t_0 = time_h.time()
            try:
                chunk = self.conn.recv(32*(1024))  # 32 KB
                if chunk != b'':
                    buff += chunk
                    loop_index_f = ((len(buff))/filesize)*self.bar_size
                    loop_index = int(loop_index_f)
                    loop_index_l = int(round(loop_index_f-loop_index, 1)*6)
                    nb_of_total = "{:.2f}/{:.2f} MB".format(
                        len(buff)/(1024**2), filesize/(1024**2))
                    percentage = len(buff)/filesize
                    tdiff = time_h.time() - t_0
                    t_elapsed = time_h.time() - t_start
                    t_speed = "{:^2.2f}".format((len(buff)/(1024**2))/t_elapsed)
                    ett = filesize / (len(buff) / t_elapsed)
                    if self.pb:
                        self.do_pg_bar(loop_index, self.wheel,
                                       nb_of_total, t_speed, t_elapsed,
                                       loop_index_l, percentage, ett)
                    else:
                        sys.stdout.write("Received %d of %d bytes\r" %
                                         (len(buff), filesize))
                        sys.stdout.flush()
                    # do_pg_bar(loop_index, wheel)
                    with open(filetoget, 'ab') as log:
                        log.write(chunk)
                    if len(buff) == filesize:
                        EOF = True
                        break
                else:
                    pass

            except Exception as e:
                if e == KeyboardInterrupt:
                    break

                else:
                    if str(e) == 'timed out':
                        break
        # print(chunk)
        # print(filesize, len(buff))
        if not EOF:
            print('CONNECTION ERROR, TRYING AGAIN...')
            self.sync_file(args)
        print('\n\nDone in {} seconds'.format(round(t_elapsed, 2)))
        print('Total data received: {:>30.2f} kB'.format(filesize/1024))
        return True

    def sync_files(self, args, dev_name):
        files_to_get = args.fre
        for file in files_to_get:
            src_file = file
            dst_file = './{}'.format(file.split('/')[-1])
            try:
                if not src_file.startswith('/'):
                    src_file = '/{}'.format(src_file)
                print("{}:{} -> {}\n".format(dev_name, src_file, dst_file))
                args.f = file
                self.sync_file(args, dev_name)
                self.dev.output_queue.get()
                print('')
            except KeyboardInterrupt as e:
                print('KeyboardInterrupt: get Operation Cancelled')
            # time_h.sleep(1)
        return True

    def disconnect(self):
        self.conn.close()
        self.dev.wr_cmd('cli_soc.close()')
        self.dev.disconnect()


def synctool(args, dev_name):
    try:
        dev = Device(args.t, args.p, init=True, autodetect=True)
        rsyncio = SyncFileIO(dev, port=8005)
        if args.m == 'fget':
            if not args.f and not args.fre:
                print('args -f or -fre required:')
                see_help(args.m)
                sys.exit()
            if not args.fre:
                if args.s is None and args.dir is None:
                    print('Looking for file in {}:/ dir ...'.format(dev_name))
                    cmd_str = "import uos;'{}' in uos.listdir()".format(args.f)
                    dir = ''
                if args.dir is not None or args.s is not None:
                    if args.s is not None:
                        args.dir = '{}/{}'.format(args.s, args.dir)
                    print('Looking for file in {}:/{} dir'.format(dev_name, args.dir))
                    cmd_str = "import uos; '{}' in uos.listdir('/{}')".format(args.f, args.dir)
                    dir = '/{}'.format(args.dir)

                try:
                    file_exists = dev.cmd(cmd_str, silent=True, rtn_resp=True)
                    if dev._traceback.decode() in dev.response:
                        try:
                            raise DeviceException(dev.response)
                        except Exception as e:
                            print(e)
                            print('Directory {}:{} does NOT exist'.format(dev_name, dir))
                            result = False
                    else:
                        if file_exists is True:
                            cmd_str = "import uos; not uos.stat('{}')[0] & 0x4000 ".format(dir + '/' + args.f)
                            is_file = dev.cmd(cmd_str, silent=True, rtn_resp=True)
                    if file_exists is True and is_file:
                        rsyncio.connect(args)
                        print('Downloading file {} @ {}...'.format(args.f, dev_name))
                        file_to_get = args.f
                        if args.s == 'sd':
                            file_to_get = '/sd/{}'.format(args.f)
                        if args.dir is not None:
                            file_to_get = '/{}/{}'.format(args.dir, args.f)
                        # if not args.wss:
                        #     copyfile_str = 'upytool -p {} {}:{} .{}'.format(
                        #         passwd, target, file_to_get, id_file)
                        dst_file = args.f
                        # if dir == '':
                        #     dir = '/'
                        src_file = file_to_get
                        if not src_file.startswith('/'):
                            src_file = '/'.join([dir, dst_file])
                        print("{}:{} -> ./{}\n".format(dev_name, src_file, dst_file))
                        args.f = file_to_get
                        result = rsyncio.sync_file(args, dev_name)
                        print('Done!')
                        rsyncio.disconnect()
                    else:
                        if file_exists is False:
                            if dir == '':
                                dir = '/'
                            print('File Not found in {}:{} directory'.format(dev_name, dir))
                        elif file_exists is True:
                            if is_file is not True:
                                print('{}:{} is a directory'.format(dev_name, dir + '/' + args.f))
                except KeyboardInterrupt as e:
                    print('KeyboardInterrupt: get Operation Cancelled')

            else:
                if args.s is None and args.dir is None:
                    print('Looking for files in {}:/ dir ...'.format(dev_name))
                    cmd_str = "import uos;[file for file in uos.listdir() if not uos.stat(file)[0] & 0x4000]"
                    dir = ''
                elif args.dir is not None or args.s is not None:
                    if args.s is not None:
                        args.dir = '{}/{}'.format(args.s, args.dir)
                    print('Looking for files in {}:/{} dir'.format(dev_name, args.dir))
                    cmd_str = "import uos;[file for file in uos.listdir('/{0}') if not uos.stat('/{0}/'+file)[0] & 0x4000]".format(args.dir)
                    dir = '/{}'.format(args.dir)
                try:
                    file_exists = dev.cmd(cmd_str, silent=True, rtn_resp=True)
                    if dev._traceback.decode() in dev.response:
                        try:
                            raise DeviceException(dev.response)
                        except Exception as e:
                            print(e)
                            print('Directory {}:{} does NOT exist'.format(dev_name, dir))
                            result = False
                    else:
                        files_to_get = []
                        # Handle special cases:
                        # CASE [cwd]:
                        if len(args.fre) == 1:
                            if args.fre[0] == 'cwd' or args.fre[0] == '.':
                                args.fre = file_exists
                            elif '*' in args.fre[0]:
                                start_exp, end_exp = args.fre[0].split('*')
                                args.fre = [file for file in file_exists if file.startswith(start_exp) and file.endswith(end_exp)]
                        if dir == '':
                            dir = '/'
                        print('Files in {}:{} to get: '.format(dev_name, dir))
                        for file in args.fre:
                            if file in file_exists:
                                files_to_get.append(file)
                                print('- {} '.format(file))
                            elif '*' in file:
                                start_exp, end_exp = file.split('*')
                                for reg_file in file_exists:
                                    if reg_file.startswith(start_exp) and reg_file.endswith(end_exp):
                                        files_to_get.append(reg_file)
                                        print('- {} '.format(reg_file))
                            else:
                                filesize = 'FileNotFoundError'
                                print('- {} [{}]'.format(file, filesize))
                        if files_to_get:
                            rsyncio.connect(args)
                            print('Downloading files @ {}...'.format(dev_name))
                            # file_to_get = args.f
                            if args.dir is not None:
                                args.fre = ['/{}/{}'.format(args.dir, file) for file in files_to_get]
                            else:
                                args.fre = files_to_get
                            result = rsyncio.sync_files(args, dev_name)
                            print('Done!')
                            rsyncio.disconnect()
                        else:
                            if dir == '':
                                dir = '/'
                            print('Files Not found in {}:{} directory'.format(dev_name, dir))
                except KeyboardInterrupt as e:
                    print('KeyboardInterrupt: get Operation Cancelled')
    except DeviceNotFound as e:
        print('ERROR {}'.format(e))

    return



UID = "from machine import unique_id;\
from ubinascii import hexlify;hexlify(unique_id());gc.collect()"

UPYSH = "from upysh import *;gc.collect()"

HELP = "help();gc.collect()"

MODULES = "help('modules');gc.collect()"

MEM = "from micropython import mem_info;mem_info();gc.collect()"

OS_STAT = "import os;os.stat('{}');gc.collect()"

FILE_STAT = "import os;[(filename,os.stat('{0}'+str(filename))[6]) for filename in os.listdir('{0}')]"

CHECK_DIR = "import os;'{}' in os.listdir('/');gc.collect()"

STAT_FS = "import os;os.statvfs('{}');gc.collect()"

IFCONFIG = "network.WLAN(network.STA_IF).ifconfig()"

SSID = "network.WLAN(network.STA_IF).config('essid')"

BSSID = "network.WLAN(network.STA_IF).config('mac')"

RSSI = "network.WLAN(network.STA_IF).status('rssi')"

NET_INFO = "import network;[{},{},{},{}];gc.collect()".format(IFCONFIG,
                                                              SSID,
                                                              BSSID,
                                                              RSSI)

NET_SCAN = "import network;network.WLAN(network.STA_IF).scan();gc.collect()"

NET_STAT_ON = "import network;network.WLAN(network.STA_IF).active(True);gc.collect()"

NET_STAT_OFF = "import network;network.WLAN(network.STA_IF).active(False);gc.collect()"

NET_STAT_CONN = "import network;network.WLAN(network.STA_IF).connect('{}', '{}');gc.collect()"

NET_STAT = "import network;network.WLAN(network.STA_IF).active();gc.collect()"

AP_ON = "import network;network.WLAN(network.AP_IF).active(True);gc.collect()"

AP_OFF = "import network;network.WLAN(network.AP_IF).active(False);gc.collect()"

AP_STATE = "network.WLAN(network.AP_IF).active()"

AP_SSID = "network.WLAN(network.AP_IF).config('essid')"

AP_CHANNEL = "network.WLAN(network.AP_IF).config('channel')"

AP_AUTHMODE = "network.WLAN(network.AP_IF).config('authmode')"

AP_IFCONFIG = "network.WLAN(network.AP_IF).ifconfig()"

APSTAT = "import network;[{},{},{},{},{}];gc.collect()".format(
    AP_STATE, AP_SSID, AP_CHANNEL, AP_AUTHMODE, AP_IFCONFIG)

AP_CONFIG = "network.WLAN(network.AP_IF).config(essid='{}',authmode=network.AUTH_WPA_WPA2_PSK, password='{}')"

AP_SCAN = "import network;network.WLAN(network.AP_IF).status('stations');gc.collect()"

I2C_CONFIG = "from machine import I2C,Pin;i2c = I2C(scl=Pin({}),sda=Pin({}));gc.collect()"

I2C_CONFIG_PYB = "from machine import I2C;i2c = I2C(scl='{}', sda='{}');gc.collect()"

I2C_SCAN = "i2c.scan();gc.collect()"

SPI_CONFIG = "from machine import SPI,Pin;spi = SPI(1, baudrate=10000000, sck=Pin({}), mosi=Pin({}), miso=Pin({})); cs = Pin({}, Pin.OUT)"

SET_RTC_LT = "from machine import RTC;rtc = RTC();rtc.datetime(({}, {}, {}, {}, {}, {}, {}+1, {}));gc.collect()"

RTC = "from machine import RTC;rtc = RTC();"

NTPTIME = "from ntptime import settime;settime();"

RTC_CONFIG = "(year, month, mday, week_of_year, hour, minute, second, milisecond) = rtc.datetime();"

UTC_ZONE = "rtc.datetime((year, month, mday, week_of_year,hour+{}, minute, second, milisecond));gc.collect()"

SET_RTC_NT = RTC + NTPTIME + RTC_CONFIG + UTC_ZONE

DATETIME = "import time; tnow = time.localtime();tnow[:6];gc.collect()"

WLAN_INIT = "from wifiutils import WIFI_UTIL; u_wlan = WIFI_UTIL(silent=False);gc.collect()"

WLAN_CONFIG = "u_wlan.sta_config('{}', '{}');gc.collect()"

WLAN_AP_CONFIG = "u_wlan.ap_config('{}', '{}');gc.collect()"

WLAN_CONN = "u_wlan.STA_conn();gc.collect()"

WLAN_AP_CONN = "u_wlan.AP_conn();gc.collect()"

SD_ENABLE_CONF = "from machine import Pin;sd_enable=Pin({}, Pin.OUT);"
SD_ENABLE_TOGGLE = "sd_enable.value(not sd_enable.value());sd_enable.value();gc.collect()"
SD_ENABLE = SD_ENABLE_CONF + SD_ENABLE_TOGGLE

SD_SDINIT = "import sdcard,os;sd = sdcard.SDCard(spi, cs);time.sleep_ms(1000);"
SD_MOUNT = "os.mount(sd, '/sd');'sd' in os.listdir('/');gc.collect()"
SD_INIT = SD_SDINIT + SD_MOUNT

SD_DEINIT = "import os;os.umount('/sd');sd_enable.off();sd_enable.value();gc.collect()"
SD_AUTO = "import SD_AM;gc.collect()"

CHECK_UPYSH2 = "import os;'upysh2.py' in os.listdir('lib');gc.collect()"

CMDDICT_ = {'UID': UID, 'UPYSH': UPYSH, 'HELP': HELP, 'MOD': MODULES,
            'MEM': MEM, 'OS_STAT': OS_STAT, 'FILE_STAT': FILE_STAT,
            'CHECK_DIR': CHECK_DIR, 'STAT_FS': STAT_FS,
            'NET_INFO': NET_INFO, 'NET_SCAN': NET_SCAN,
            'NET_STAT_ON': NET_STAT_ON, 'NET_STAT_OFF': NET_STAT_OFF,
            'NET_STAT_CONN': NET_STAT_CONN, 'NET_STAT': NET_STAT,
            'AP_ON': AP_ON, 'AP_OFF': AP_OFF, 'APSTAT': APSTAT,
            'AP_CONFIG': AP_CONFIG, 'AP_SCAN': AP_SCAN,
            'I2C_CONFIG': I2C_CONFIG, 'I2C_SCAN': I2C_SCAN,
            'SPI_CONFIG': SPI_CONFIG, 'SET_RTC_LT': SET_RTC_LT,
            'SET_RTC_NT': SET_RTC_NT, 'DATETIME': DATETIME,
            'I2C_CONFIG_PYB': I2C_CONFIG_PYB, 'WLAN_INIT': WLAN_INIT,
            'WLAN_CONFIG': WLAN_CONFIG, 'WLAN_AP_CONFIG': WLAN_AP_CONFIG,
            'WLAN_CONN': WLAN_CONN, 'WLAN_AP_CONN': WLAN_AP_CONN,
            'SD_ENABLE': SD_ENABLE, 'SD_INIT': SD_INIT,
            'SD_DEINIT': SD_DEINIT, 'SD_AUTO': SD_AUTO, 'CHECK_UPYSH2': CHECK_UPYSH2}

_CMDDICT_ = {k: 'import gc;' + v for k, v in CMDDICT_.items()}





GROUP_MODE_HELP = """
> GROUP COMMAND MODE:
    To send a command to multiple devices in a group (made with make_group command)

    To target specific devices within a group add -devs option as -devs [DEV NAME] [DEV NAME] ...

    *(upydev will use local working directory configuration unless it does
    not find any or manually indicated with -g option)

    COMMAND MODE OPTION:
        -G : Usage '$ upydev ACTION -G GROUPNAME [opts]' or
                   '$ upydev ACTION -gg [opts]' for global group.
            This sends the command to one device at a time;

        -GP: Usage '$ upydev ACTION -GP GROUPNAME [opts]' or
                   '$ upydev ACTION -ggp [opts]' for global group.
            For parallel/non-blocking command execution using multiprocessing"""



import os
import time
import sys
from datetime import timedelta
from upydevice import Device, DeviceNotFound, DeviceException
from upydev.helpinfo import see_help
from upydev.dsyncio import d_sync_recursive
from upydev import upip_host
import shutil
import glob


class BleFileIO:
    def __init__(self, dev, port=None):
        self.host = None
        self.port = port
        self.dev = dev
        self.buff = bytearray(1024*2)
        self.bloc_progress = ["▏", "▎", "▍", "▌", "▋", "▊", "▉"]
        self.columns, self.rows = os.get_terminal_size(0)
        self.cnt_size = 65
        self.bar_size = int((self.columns - self.cnt_size))
        self.pb = False
        self.wheel = ['|', '/', '-', "\\"]

    def get_pb(self):
        self.columns, self.rows = os.get_terminal_size(0)
        if self.columns > self.cnt_size:
            self.bar_size = int((self.columns - self.cnt_size))
            self.pb = True
        else:
            self.bar_size = 1
            self.pb = False

    def do_pg_bar(self, index, wheel, nb_of_total, speed, time_e, loop_l,
                  percentage, ett):
        l_bloc = self.bloc_progress[loop_l]
        if index == self.bar_size:
            l_bloc = "█"
        sys.stdout.write("\033[K")
        print('▏{}▏{:>2}{:>5} % | {} | {:>5} KB/s | {}/{} s'.format("█" *index + l_bloc  + " "*((self.bar_size+1) - len("█" *index + l_bloc)),
                                                                        wheel[index % 4],
                                                                        int((percentage)*100),
                                                                        nb_of_total, speed,
                                                                        str(timedelta(seconds=time_e)).split('.')[0][2:],
                                                                        str(timedelta(seconds=ett)).split('.')[0][2:]), end='\r')
        sys.stdout.flush()

    def get(self, src, dst_file, chunk_size=256, ppath=False, dev_name=None):  # from Pyboard.py
        self.get_pb()
        cnt = 0
        t_start = time.time()
        # def_chunk = chunk
        sz = self.dev.cmd("import uos; uos.stat('{}')[6]".format(src), silent=True,
                          rtn_resp=True)

        dst_file = src.split('/')[-1]
        if ppath:
            print('{}:{} -> {}'.format(dev_name, src, dst_file), end='\n\n')
        print("{}  [{:.2f} KB]".format(src, sz / 1024))
        self.dev.flush()
        # self.start_SOC()
        self.dev.cmd("f=open('%s','rb');r=f.read" % src, silent=True)
        with open(dst_file, 'wb') as f:
            pass
        with open(dst_file, 'ab') as f:
            while True:
                data = b''
                data = self.dev.cmd("print(r(%u))" % chunk_size, rtn_resp=True,
                                    silent=True)
                if data == b'':
                    break
                f.write(data)
                cnt += len(data)
                loop_index_f = (cnt/sz)*self.bar_size
                loop_index = int(loop_index_f)
                loop_index_l = int(round(loop_index_f-loop_index, 1)*6)
                nb_of_total = "{:.2f}/{:.2f} KB".format(cnt/(1024), sz/(1024))
                percentage = cnt / sz
                t_elapsed = time.time() - t_start
                t_speed = "{:^2.2f}".format((cnt/(1024))/t_elapsed)
                ett = sz / (cnt / t_elapsed)
                if self.pb:
                    self.do_pg_bar(loop_index, self.wheel,
                                   nb_of_total, t_speed, t_elapsed,
                                   loop_index_l, percentage, ett)
        print('\n')
        self.dev.cmd("f.close()")
        return True

    def get_files(self, args, dev_name):
        files_to_get = args.fre
        for file in files_to_get:
            src_file = file
            dst_file = './{}'.format(file.split('/')[-1])
            try:
                if not src_file.startswith('/'):
                    src_file = '/{}'.format(src_file)
                print("{}:{} -> {}\n".format(dev_name, src_file, dst_file))
                self.get(file, dst_file)
            except KeyboardInterrupt as e:
                print('KeyboardInterrupt: get Operation Cancelled')
                self.dev.cmd("f.close()", silent=True)
        return True

    def put(self, src, dst_file, chunk_size=256, abs_path=True, ppath=False,
            dev_name=None):  # from Pyboard.py
        self.get_pb()
        sz = os.stat(src)[6]
        cnt = 0
        t_start = time.time()
        if not abs_path:
            src_ori = src
            src = src.split('/')[-1]
        if ppath:
            if not dst_file.startswith('/'):
                print("{} -> {}:/{}\n".format(src, dev_name, dst_file))
            else:
                print("{} -> {}:{}\n".format(src, dev_name, dst_file))
        print("{}  [{:.2f} KB]".format(src, sz / 1024))
        self.dev.cmd("f=open('%s','wb');w=f.write" % dst_file, silent=True)
        if not abs_path:
            src = src_ori
        with open(src, 'rb') as f:
            while True:
                data = f.read(chunk_size)
                if not data:
                    break

                if len(data) > self.dev.len_buffer:
                    for i in range(0, len(data), self.dev.len_buffer):
                        self.dev.cmd('w(' + repr(data[i:i+self.dev.len_buffer]) + ')',
                                     silent=True)
                else:
                    self.dev.cmd('w(' + repr(data) + ')', silent=True)

                cnt += len(data)
                loop_index_f = (cnt/sz)*self.bar_size
                loop_index = int(loop_index_f)
                loop_index_l = int(round(loop_index_f-loop_index, 1)*6)
                nb_of_total = "{:.2f}/{:.2f} KB".format(cnt/(1024), sz/(1024))
                percentage = cnt / sz
                t_elapsed = time.time() - t_start
                t_speed = "{:^2.2f}".format((cnt/(1024))/t_elapsed)
                ett = sz / (cnt / t_elapsed)
                if self.pb:
                    self.do_pg_bar(loop_index, self.wheel,
                                   nb_of_total, t_speed, t_elapsed,
                                   loop_index_l, percentage, ett)
        print('\n')
        self.dev.cmd("f.close()", silent=True)
        self.dev.flush()
        return True

    def put_files(self, args, dev_name):
        files_to_put = args.fre
        for file in files_to_put:
            src_file = file
            dst_file = file.split('/')[-1]
            if args.s:
                if args.s.startswith('./'):
                    args.s = args.s.replace('./', '')
                if args.s != '/':
                    dst_file = '/'.join([args.s, dst_file])
                else:
                    dst_file = ''.join([args.s, dst_file])
            try:
                abs_dst_file = dst_file
                if not dst_file.startswith('/'):
                    abs_dst_file = '/{}'.format(dst_file)
                print("{} -> {}:{}\n".format(src_file, dev_name, abs_dst_file))
                self.put(src_file, dst_file)
            except KeyboardInterrupt as e:
                print('KeyboardInterrupt: put Operation Cancelled')
                self.dev.cmd("f.close()", silent=True)
        return True

    def upip_install(self, args, dev_name):
        try:
            dir_lib = 'lib'
            lib = args.f
            dev_lib = args.dir
            if not dev_lib:
                dev_lib = "./"
            pckg_content, pckg_dir = upip_host.install_pkg(lib, ".")
            # sync local lib to device lib
            print('Installing {} to {}:/lib'.format(pckg_dir, dev_name))
            # cwd_now = self.dev.cmd('os.getcwd()', silent=True, rtn_resp=True)
            # if self.dev.dev_platform == 'pyboard':
            # self.dev.cmd("os.chdir('/flash')")
            # d_sync_recursive(dir_lib, show_tree=True, root_sync_folder=".")
            d_sync_recursive(dir_lib, devIO=self,
                             show_tree=True,
                             rootdir=dev_lib,
                             root_sync_folder=dir_lib,
                             args=args,
                             dev_name=dev_name)
            rm_lib = input('Do you want to remove local lib? (y/n): ')
            if rm_lib == 'y':
                shutil.rmtree(dir_lib)
            print("Successfully installed {} to {} :/lib".format(pckg_dir, dev_name))
            # self.dev.cmd("os.chdir('{}')".format(cwd_now))
        except Exception as e:
            print('Please indicate a library to install')


def bletool(args, dev_name):
    if not args.f and not args.fre:
        print('args -f or -fre required:')
        see_help(args.m)
        sys.exit()
    try:
        dev = Device(args.t, args.p, init=True)
        bleio = BleFileIO(dev)
        if args.m == 'put':
            if not args.f and not args.fre:
                print('args -f or -fre required:')
                see_help(args.m)
                sys.exit()
            if not args.fre:
                # One file:
                source = ''
                if args.s == 'sd':
                    source += '/' + args.s
                file = args.f
                if args.dir is not None:
                    source += '/' + args.dir
                file_in_upy = file.split('/')[-1]
                if source:
                    args.s = source + '/'
                # Check if file exists:
                try:
                    os.stat(file)[6]
                    if not os.path.isdir(file):
                        if args.s:
                            is_dir_cmd = "import uos, gc; uos.stat('{}')[0] & 0x4000".format(source)
                            is_dir = dev.cmd(is_dir_cmd, silent=True, rtn_resp=True)
                            # dev.disconnect()
                            if dev._traceback.decode() in dev.response:
                                try:
                                    raise DeviceException(dev.response)
                                except Exception as e:
                                    print(e)
                                    print('Directory {}:{} does NOT exist'.format(dev_name, source))
                                    result = False
                            else:
                                if is_dir:
                                    print('Uploading file {} @ {}...'.format(file_in_upy, dev_name))
                                    src_file = file
                                    dst_file = args.s + file_in_upy
                                    print("{} -> {}:{}\n".format(src_file, dev_name, dst_file))
                                    result = bleio.put(file, dst_file)
                        else:
                            print('Uploading file {} @ {}...'.format(file_in_upy, dev_name))
                            src_file = file
                            if not source:
                                source = '/'
                            dst_file = source + file_in_upy
                            print("{} -> {}:{}\n".format(src_file, dev_name, dst_file))
                            result = bleio.put(file, file_in_upy)

                        # Reset:
                        if result:
                            if args.rst is None:
                                time.sleep(0.4)
                                dev.reset(reconnect=False)
                                time.sleep(0.4)
                    else:
                        print('FileNotFoundError: {} is a directory'.format(file))
                except FileNotFoundError as e:
                    print('FileNotFoundError:', e)
                except KeyboardInterrupt as e:
                    print('KeyboardInterrupt: put Operation Cancelled')
            else:
                # Handle special cases:
                # CASE [cwd]:
                if args.fre[0] == 'cwd' or args.fre[0] == '.':
                    args.fre = [fname for fname in os.listdir('./') if os.path.isfile(fname) and not fname.startswith('.')]
                    print('Files in ./{} to put: '.format(os.getcwd().split('/')[-1]))

                elif '*' in args.fre[0]:
                    args.fre = glob.glob(args.fre[0])
                    print('Files to put: ')

                else:
                    print('Files to put: ')

                files_to_put = []
                for file in args.fre:
                    try:
                        filesize = os.stat(file)[6]
                        if not os.path.isdir(file):
                            files_to_put.append(file)
                            print('- {} [{:.2f} KB]'.format(file, filesize/1024))
                        else:
                            filesize = 'IsDirectory'
                            print('- {} [{}]'.format(file, filesize))
                    except Exception as e:
                        filesize = 'FileNotFoundError'
                        print('- {} [{}]'.format(file, filesize))
                args.fre = files_to_put
                # Multiple file:
                source = ''
                if args.s == 'sd':
                    source += '/' + args.s
                file = ''
                if args.dir is not None:
                    source += '/' + args.dir
                file_in_upy = file.split('/')[-1]
                if source:
                    args.s = source
                try:
                    if args.s:
                        is_dir_cmd = "import uos, gc; uos.stat('{}')[0] & 0x4000".format(source)
                        is_dir = dev.cmd(is_dir_cmd, silent=True, rtn_resp=True)
                        if dev._traceback.decode() in dev.response:
                            try:
                                raise DeviceException(dev.response)
                            except Exception as e:
                                print(e)
                                print('Directory {}:{} does NOT exist'.format(dev_name, source))
                                result = False
                        else:
                            if is_dir:
                                print('\nUploading files @ {}...\n'.format(dev_name))
                                result = bleio.put_files(args, dev_name)
                    else:
                        args.s = source
                        print('\nUploading files @ {}...\n'.format(dev_name))
                        result = bleio.put_files(args, dev_name)

                    # Reset:
                    if result:
                        if args.rst is None:
                            time.sleep(0.4)
                            dev.reset(reconnect=False)
                            time.sleep(0.4)
                            # dev.disconnect()
                except KeyboardInterrupt as e:
                    print('KeyboardInterrupt: put Operation Cancelled')
        elif args.m == 'get':
            if not args.f and not args.fre:
                print('args -f or -fre required:')
                see_help(args.m)
                sys.exit()
            if not args.fre:
                if args.s is None and args.dir is None:
                    print('Looking for file in {}:/ dir ...'.format(dev_name))
                    cmd_str = "import uos;'{}' in uos.listdir()".format(args.f)
                    dir = ''
                if args.dir is not None or args.s is not None:
                    if args.s is not None:
                        args.dir = '{}/{}'.format(args.s, args.dir)
                    print('Looking for file in {}:/{} dir'.format(dev_name, args.dir))
                    cmd_str = "import uos; '{}' in uos.listdir('/{}')".format(args.f, args.dir)
                    dir = '/{}'.format(args.dir)

                try:
                    file_exists = dev.cmd(cmd_str, silent=True, rtn_resp=True)
                    if dev._traceback.decode() in dev.response:
                        try:
                            raise DeviceException(dev.response)
                        except Exception as e:
                            print(e)
                            print('Directory {}:{} does NOT exist'.format(dev_name, dir))
                            result = False
                    else:
                        if file_exists is True:
                            cmd_str = "import uos; not uos.stat('{}')[0] & 0x4000 ".format(dir + '/' + args.f)
                            is_file = dev.cmd(cmd_str, silent=True, rtn_resp=True)
                    if file_exists is True and is_file:
                        print('Downloading file {} @ {}...'.format(args.f, dev_name))
                        file_to_get = args.f
                        if args.s == 'sd':
                            file_to_get = '/sd/{}'.format(args.f)
                        if args.dir is not None:
                            file_to_get = '/{}/{}'.format(args.dir, args.f)
                        # if not args.wss:
                        #     copyfile_str = 'upytool -p {} {}:{} .{}'.format(
                        #         passwd, target, file_to_get, id_file)
                        dst_file = args.f
                        # if dir == '':
                        #     dir = '/'
                        src_file = file_to_get
                        if not src_file.startswith('/'):
                            src_file = '/'.join([dir, dst_file])
                        print("{}:{} -> ./{}\n".format(dev_name, src_file, dst_file))
                        result = bleio.get(file_to_get, dst_file)
                        print('Done!')
                    else:
                        if file_exists is False:
                            if dir == '':
                                dir = '/'
                            print('File Not found in {}:{} directory'.format(dev_name, dir))
                        elif file_exists is True:
                            if is_file is not True:
                                print('{}:{} is a directory'.format(dev_name, dir + '/' + args.f))
                except KeyboardInterrupt as e:
                    print('KeyboardInterrupt: get Operation Cancelled')

            else:
                if args.s is None and args.dir is None:
                    print('Looking for files in {}:/ dir ...'.format(dev_name))
                    cmd_str = "import uos;[file for file in uos.listdir() if not uos.stat(file)[0] & 0x4000]"
                    dir = ''
                elif args.dir is not None or args.s is not None:
                    if args.s is not None:
                        args.dir = '{}/{}'.format(args.s, args.dir)
                    print('Looking for files in {}:/{} dir'.format(dev_name, args.dir))
                    cmd_str = "import uos;[file for file in uos.listdir('/{0}') if not uos.stat('/{0}/'+file)[0] & 0x4000]".format(args.dir)
                    dir = '/{}'.format(args.dir)
                try:
                    file_exists = dev.cmd(cmd_str, silent=True, rtn_resp=True)
                    if dev._traceback.decode() in dev.response:
                        try:
                            raise DeviceException(dev.response)
                        except Exception as e:
                            print(e)
                            print('Directory {}:{} does NOT exist'.format(dev_name, dir))
                            result = False
                    else:
                        files_to_get = []
                        # Handle special cases:
                        # CASE [cwd]:
                        if len(args.fre) == 1:
                            if args.fre[0] == 'cwd' or args.fre[0] == '.':
                                args.fre = file_exists
                            elif '*' in args.fre[0]:
                                start_exp, end_exp = args.fre[0].split('*')
                                args.fre = [file for file in file_exists if file.startswith(start_exp) and file.endswith(end_exp)]
                        if dir == '':
                            dir = '/'
                        print('Files in {}:{} to get: '.format(dev_name, dir))
                        for file in args.fre:
                            if file in file_exists:
                                files_to_get.append(file)
                                print('- {} '.format(file))
                            elif '*' in file:
                                start_exp, end_exp = file.split('*')
                                for reg_file in file_exists:
                                    if reg_file.startswith(start_exp) and reg_file.endswith(end_exp):
                                        files_to_get.append(reg_file)
                                        print('- {} '.format(reg_file))
                            else:
                                filesize = 'FileNotFoundError'
                                print('- {} [{}]'.format(file, filesize))
                        if files_to_get:
                            print('Downloading files @ {}...'.format(dev_name))
                            # file_to_get = args.f
                            if args.dir is not None:
                                args.fre = ['/{}/{}'.format(args.dir, file) for file in files_to_get]
                            else:
                                args.fre = files_to_get
                            result = bleio.get_files(args, dev_name)
                            print('Done!')
                        else:
                            if dir == '':
                                dir = '/'
                            print('Files Not found in {}:{} directory'.format(dev_name, dir))
                except KeyboardInterrupt as e:
                    print('KeyboardInterrupt: get Operation Cancelled')
    except DeviceNotFound as e:
        print('ERROR {}'.format(e))

    if dev.is_connected():
        dev.disconnect()
        time.sleep(0.1)

    return



#!/usr/bin/env python3

import os
import time
import json
import hashlib
from binascii import hexlify


# AVOID FILES

_CONFG_FILES = ['.upydev_wdlog.json', 'upydev_.config']


# from @Roberthh #https://forum.micropython.org/viewtopic.php?f=2&t=7512
def rmrf(d):  # Remove file or tree
    try:
        if os.path.isdir(d):  # Dir
            for f in os.listdir(d):
                if f not in ('.', '..'):
                    rmrf(os.path.join(d, f))  # File or Dir
            os.rmdir(d)
        else:  # File
            os.remove(d)
    except Exception as e:
        print(e)
        print("rm of '%s' failed" % d)


# WATCHDOG LOG FILE

def get_hash(file):
    with open(file, 'rb') as file_to_hash:
        raw_file = file_to_hash.read()
    file_hash = hashlib.sha256()
    file_hash.update(raw_file)
    hashed_file = file_hash.digest()
    return hexlify(hashed_file).decode()


def get_hash_cwd_dict(path='.'):
    hash_cwd_dict = {name: get_hash(os.path.join(path, name)) for name in
                     [file for file in os.listdir(path) if
                      os.path.isfile(os.path.join(path, file))
                      and file not in _CONFG_FILES]}
    return hash_cwd_dict


def check_wdlog(path='.', save_wdlog=True):
    # WD_LOG EXISTS, CHECK AND COMPARE
    deleted_files = []
    if '.upydev_wdlog.json' in os.listdir(path):
        print('Checking upydev cwd watchdog logfile...')
        with open('{}/'.format(path) + '.upydev_wdlog.json', 'r') as wd_logfile:
            hash_wdlog_dict = json.loads(wd_logfile.read())
            files_wdlog_list = list(hash_wdlog_dict.keys())
            files_cwd = [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file)) and file not in _CONFG_FILES]
            # New files in cwd:
            new_files_to_upload = [file for file in files_cwd if file not in files_wdlog_list]
            if len(new_files_to_upload) > 0:
                print('New files to upload:')
                for nf in new_files_to_upload:
                    print('- {}'.format(nf))
            # Files deleted in cwd:
            deleted_files = [file for file in files_wdlog_list if file not in files_cwd]
            # Files modified in cwd:
            files_wdlog_list = [file for file in files_wdlog_list if file not in deleted_files]
            modified_files = [file for file in files_wdlog_list if get_hash(os.path.join(path, file)) != hash_wdlog_dict[file]]
            if len(modified_files) > 0:
                print('Modified files to upload:')
                for mf in modified_files:
                    print('- {}'.format(mf))
            if len(deleted_files) > 0:
                print('Deleted files to remove:')
                for df in deleted_files:
                    print('- {}'.format(df))
            # MAKE NEW WD_LOG
            if save_wdlog:
                with open('{}/'.format(path) + '.upydev_wdlog.json', 'w') as wd_logfile:
                    hash_cwd_dict = get_hash_cwd_dict(path=path)
                    wd_logfile.write(json.dumps(hash_cwd_dict))
            global_files_to_upload = new_files_to_upload + modified_files
            if len(global_files_to_upload) == 0:
                print('No new or modified files found')
            return global_files_to_upload, deleted_files
    # WD_LOG does NOT exist, CREATE NEW ONE
    else:
        if save_wdlog:
            print('.upydev_wdlog.json not found, creating new one...')
            with open('{}/'.format(path) + '.upydev_wdlog.json', 'w') as wd_logfile:
                hash_cwd_dict = get_hash_cwd_dict(path=path)
                wd_logfile.write(json.dumps(hash_cwd_dict))
            print('Done!')
        global_files_to_upload = [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file)) and file not in _CONFG_FILES]
        for file in global_files_to_upload:
            print('- {}'.format(file))
        return global_files_to_upload, deleted_files


class LTREE:

    def __repr__(self):
        self.__call__()
        return ""

    def __call__(self, path=".", level=0, is_last=False, is_root=True,
                 carrier="│   ", hidden=False):
        if is_root:
            print('\u001b[34;1m{}\033[0m'.format(path))
        os.chdir(path)
        r_path = path
        path = "."
        if hidden:
            l = os.listdir(path)
        else:
            l = [f for f in os.listdir(path) if not f.startswith('.')]
        nf = len([file for file in l if not os.stat("%s/%s" % (path, file))[0] & 0x4000])
        nd = len(l) - nf
        ns_f, ns_d = 0, 0
        l.sort()
        if len(l) > 0:
            last_file = l[-1]
        else:
            last_file = ''
        for f in l:
            st = os.stat("%s/%s" % (path, f))
            if st[0] & 0x4000:  # stat.S_IFDIR
                print(self._treeindent(level, f, last_file, is_last=is_last, carrier=carrier) + "  \u001b[34;1m%s\033[0m" % f)
                if f == last_file and level == 0:
                    carrier = "    "
                os.chdir(f)
                level += 1
                lf = last_file == f
                if level > 1:
                    if lf:
                        carrier += "     "
                    else:
                        carrier += "    │"
                ns_f, ns_d = self.__call__(level=level, is_last=lf,
                                           is_root=False, carrier=carrier,
                                           hidden=hidden)
                if level > 1:
                    carrier = carrier[:-5]
                os.chdir('..')
                level += (-1)
                nf += ns_f
                nd += ns_d
            else:
                print(self._treeindent(level, f, last_file, is_last=is_last, carrier=carrier) + " %s" % (f))
        if is_root:
            nd_str = 'directories'
            nf_str = 'files'
            if nd == 1:
                nd_str = 'directory'
            if nf == 1:
                nf_str = 'file'
            print('\n{} {}, {} {}'.format(nd, nd_str, nf, nf_str))
            if r_path != ".":
                os.chdir('..')
        else:
            return (nf, nd)

    def _treeindent(self, lev, f, lastfile, is_last=False, carrier=None):
        if lev == 0:
            if f != lastfile:
                return "├──"
            else:
                return "└──"
        else:
            if f != lastfile:
                return carrier + "    ├────"
            else:
                return carrier + "    └────"


tree = LTREE()


class DISK_USAGE:

    def __repr__(self):
        self.__call__()
        return ""

    def __call__(self, path=".", dlev=0, max_dlev=0, hidden=False, absp=True):
        if path != ".":
                if not os.stat(path)[0] & 0x4000:
                    print('{:9} {}'.format(self.print_filesys_info(os.stat(path)[6]), path))
                else:
                    if hidden:
                        resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path)}
                    else:
                        resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path) if not dir.startswith('.')}
                    for dir in resp.keys():

                        if not os.stat(dir)[0] & 0x4000:
                            if absp:
                                print('{:9} {}'.format(self.print_filesys_info(resp[dir]), dir))
                            else:
                                print('{:9} {}'.format(self.print_filesys_info(resp[dir]), dir.split('/')[-1]))

                        else:
                            if dlev < max_dlev:
                                dlev += 1
                                self.__call__(path=dir, dlev=dlev, max_dlev=max_dlev, hidden=hidden)
                                dlev += (-1)
                            else:
                                if absp:
                                    print('{:9} \u001b[34;1m{}\033[0m'.format(self.print_filesys_info(self.get_dir_size_recursive(dir)), dir))
                                else:
                                    print('{:9} \u001b[34;1m{}\033[0m'.format(self.print_filesys_info(self.get_dir_size_recursive(dir)), dir.split('/')[-1]))

        else:
            if hidden:
                resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path)}
            else:
                resp = {path+'/'+dir: os.stat(path+'/'+dir)[6] for dir in os.listdir(path) if not dir.startswith('.')}
            for dir in resp.keys():

                if not os.stat(dir)[0] & 0x4000:
                    print('{:9} {}'.format(self.print_filesys_info(resp[dir]), dir))

                else:
                    if dlev < max_dlev:
                        dlev += 1
                        self.__call__(path=dir, dlev=dlev, max_dlev=max_dlev, hidden=hidden)
                        dlev += (-1)
                    else:
                        print('{:9} \u001b[34;1m{}\033[0m'.format(self.print_filesys_info(self.get_dir_size_recursive(dir)), dir))

    def print_filesys_info(self, filesize):
        _kB = 1024
        if filesize < _kB:
            sizestr = str(filesize) + " by"
        elif filesize < _kB**2:
            sizestr = "%0.1f KB" % (filesize / _kB)
        elif filesize < _kB**3:
            sizestr = "%0.1f MB" % (filesize / _kB**2)
        else:
            sizestr = "%0.1f GB" % (filesize / _kB**3)
        return sizestr

    def get_dir_size_recursive(self, dir):
        return sum([os.stat(dir+'/'+f)[6] if not os.stat(dir+'/'+f)[0] & 0x4000 else self.get_dir_size_recursive(dir+'/'+f) for f in os.listdir(dir)])


du = DISK_USAGE()


def print_filesys_info(filesize):
    _kB = 1024
    if filesize < _kB:
        sizestr = str(filesize) + " by"
    elif filesize < _kB**2:
        sizestr = "%0.1f KB" % (filesize / _kB)
    elif filesize < _kB**3:
        sizestr = "%0.1f MB" % (filesize / _kB**2)
    else:
        sizestr = "%0.1f GB" % (filesize / _kB**3)
    return sizestr


def get_dir_size_recursive(dir):
    return sum([os.stat(dir+'/'+f)[6] if not os.stat(dir+'/'+f)[0] &
                0x4000 else get_dir_size_recursive(dir+'/'+f) for f in
                os.listdir(dir)])


def d_sync_recursive(folder, devIO=None, rootdir='./', root_sync_folder=None,
                     show_tree=False, args=None, dev_name=None):
    t0 = time.time()
    # type_file_dict = {True: '<f>', False: '<d>'}
    if folder == root_sync_folder:
        print('DIRECTORY TO SYNC: {}'.format(folder))
        print('DIRECTORY SIZE: {}'.format(
                                          print_filesys_info(
                                              get_dir_size_recursive(folder))))
    if show_tree:
        print('DIRECTORY TREE STRUCTURE:\n')
        tree(path=folder)
    time.sleep(1)
    print('\n')
    print('***** {} *****'.format(folder))
    print('\n')
    directory = folder
    file_list = None
    file_list_abs_path = []
    dir_list_abs_path = []
    current_dir = directory
    # get directory structure:
    print('ROOT DIRECTORY: {}'.format(rootdir))
    print('DIRECTORY TO SYNC: {}'.format(directory))
    print('\n')
    if directory != '.':
        print('CHECKING IF DIRECTORY {} IN: {}'.format(
            directory.split('/')[-1], rootdir))
    if directory.split('/')[-1] in os.listdir(rootdir) or directory == '.':
        print('DIRECTORY {} FOUND'.format(directory))
        print('\n')
        print('FILES/DIRS IN DIRECTORY {}:'.format(directory))
        if directory == '.':
            directory = ''
        du(path='./{}'.format(directory), absp=False, hidden=True)
        # for file in os.listdir(directory):
        #     print('- {} {} [{}]'.format(type_file_dict[os.path.isfile(os.path.join(current_dir, file))],
        #                            file, print_filesys_info(os.stat(os.path.join(current_dir, file))[6])))
        if directory == '':
            directory = '.'
        file_list = os.listdir(directory)
        print('\n')
        if args.wdl:
            modified_files, deleted_files = check_wdlog(path=directory)
        for file in file_list:
            if os.path.isfile(os.path.join(current_dir, file)):
                if args.wdl:
                    wdl_file = file.split('/')[-1]
                    if wdl_file not in _CONFG_FILES and wdl_file in modified_files:
                        file_list_abs_path.append(os.path.join(current_dir, file))
                else:
                    if file not in _CONFG_FILES:
                        file_list_abs_path.append(os.path.join(current_dir, file))
            elif os.path.isdir(os.path.join(current_dir, file)):
                dir_list_abs_path.append(os.path.join(current_dir, file))

    if file_list_abs_path:
        print('LIST OF FILES TO UPLOAD:')
        for file in file_list_abs_path:
            print('- {}'.format(file.split('/')[-1]))
        print('\n')
    else:
        pass
        # print('NO FILES TO UPLOAD')

    if dir_list_abs_path:
        print('LIST OF SUBDIRS TO CREATE:')
        for subdir in dir_list_abs_path:
            print('- {}'.format(subdir.split('/')[-1]))
        print('\n')
    else:
        # print('NO SUBDIRS TO CREATE')
        pass
    # Now create the root sync dir:
    if './' == rootdir:
        rootdir = ''
    if not devIO.dev.connected:
        devIO.dev.connect()
    dev_root_list = devIO.dev.wr_cmd("import os;'{}' in os.listdir('{}')".format(current_dir.split('/')[-1],
                                     rootdir), silent=True, rtn_resp=True)
    if not dev_root_list and current_dir != '.':
        print('\n')
        print('MAKING DIR: {}'.format(current_dir))
        print('\n')
        devIO.dev.wr_cmd("os.mkdir('{}')".format(current_dir), silent=True)
        print('\n')

    if file_list_abs_path:
        print('UPLOADING FILES TO {}'.format(current_dir))
        if len(file_list_abs_path) > 1:
            if directory == '.':
                file_list_abs_path = [file.split('/')[-1] for file in file_list_abs_path]
            for file in file_list_abs_path:
                print('- {}'.format(file))
            args.fre = file_list_abs_path
            if directory != '.':
                args.s = os.path.join(*file_list_abs_path[0].split('/')[:-1])
            else:
                args.s = '/'
            print('\n')
            devIO.put_files(args, dev_name)
            args.fre = None
            time.sleep(0.2)
        elif len(file_list_abs_path) == 1:
            args.fre = None
            file_to_put = file_list_abs_path[0]
            if directory == '.':
                file_to_put = file.split('/')[-1]
            print('- {}'.format(file_to_put), end='\n\n')
            file_to_put_in_dev = file_to_put.replace('./', '')
            devIO.put(file_to_put, file_to_put_in_dev, ppath=True, dev_name=dev_name)
            time.sleep(0.2)
    else:
        print('NO FILES TO UPLOAD')
    # Now create subdirs:
    print('\n')
    if dir_list_abs_path:
        print('MAKING DIRS NOW...')
        for dir_ in dir_list_abs_path:
            print('\n')
            current_dir = dir_
            # Now create the root sync dir:
            try:
                dir_to_sync = directory
                dir_to_find = current_dir.split('/')[-1]
                dev_directory = devIO.dev.cmd("'{}' in os.listdir('{}')".format(dir_to_find, dir_to_sync),
                                              silent=True, rtn_resp=True)
            except Exception as e:
                pass
            if not dev_directory:
                print('Creating dir: {}'.format(current_dir))
                devIO.dev.cmd("os.mkdir('{}')".format(current_dir))
            else:
                print('DIRECTORY {} ALREADY EXISTS'.format(dir_to_find))
    else:
        print('NO DIRS TO MAKE')

    if args.rf:
        # remove dirs and files? -rf?
        # Deleted Files
        if args.wdl:
            if deleted_files:
                print('FILES TO REMOVE:')
                for dfile in deleted_files:
                    try:
                        devIO.dev.cmd("os.remove('{}/{}')".format(directory, dfile), silent=True, rtn_resp=True)
                        if directory != '.':
                            print('- {}/{}'.format(directory, dfile))
                        else:
                            print('- {}'.format(dfile))
                    except Exception as e:
                        print(e)
                print('\nFILES DELETED')

        else:
            pass

        # Deleted Directories

        if directory != '.':
            dirs_in_dev = devIO.dev.cmd("['{0}/'+dir for dir in os.listdir('{0}') if os.stat('{0}/'+dir)[0] & 0x4000]".format(directory),
                                        silent=True, rtn_resp=True)

            deleted_dirs = [dir for dir in dirs_in_dev if dir not in dir_list_abs_path]

            if deleted_dirs:
                print('\nDIRS TO REMOVE:')
                for ddir in deleted_dirs:
                    print('- {} '.format(ddir))
                    is_dir_empty = devIO.dev.cmd("from upysh2 import rmrf;not len(os.listdir('{}')) > 0".format(ddir),
                                                 silent=True, rtn_resp=True)
                    if is_dir_empty:
                        devIO.dev.cmd("os.rmdir('{}')".format(ddir), silent=True,
                                      rtn_resp=True)
                    else:
                        devIO.dev.cmd("rmrf('{}')".format(ddir), silent=True,
                                      rtn_resp=True)


    root = directory
    for dir_ in dir_list_abs_path:
        d_sync_recursive(dir_, devIO, root, args=args, dev_name=dev_name)

    if directory == root_sync_folder:
        print('Done in : {:.2f} seconds'.format(time.time()-t0))


def dev2host_sync_recursive(folder, devIO=None, rootdir='./', root_sync_folder=None,
                            show_tree=False, args=None, dev_name=None):
    t0 = time.time()
    # type_file_dict = {True: '<f>', False: '<d>'}
    if not devIO.dev.connected:
        devIO.dev.connect()
    devIO.dev.wr_cmd('from upysh2 import du', silent=True)
    dev_dir_size = devIO.dev.wr_cmd("du.get_dir_size_recursive('{}')".format(folder),
                                    silent=True, rtn_resp=True)
    _folder = folder if folder != '.' else ''
    if folder == root_sync_folder:
        print('DIRECTORY TO SYNC: {}:/{}'.format(dev_name, _folder))
        print('DIRECTORY SIZE: {}'.format(
                                          print_filesys_info(dev_dir_size)))
        platform = devIO.dev.wr_cmd('import os; os.uname().sysname', silent=True,
                                    rtn_resp=True)
        # print(platform)
        if platform == 'pyboard':
            rootdir = '.'
    if show_tree:
        devIO.dev.wr_cmd('from upysh2 import tree', silent=True)
        print('DIRECTORY TREE STRUCTURE:\n')
        devIO.dev.wr_cmd("tree('{}')".format(folder), follow=True)
        # tree(path=folder)
    time.sleep(1)
    print('\n')
    print('***** {}:/{} *****'.format(dev_name, _folder))
    print('\n')
    directory = folder
    file_list = None
    file_list_abs_path = []
    dir_list_abs_path = []
    current_dir = directory
    # get directory structure:
    print('ROOT DIRECTORY: {}:{}'.format(dev_name, rootdir))
    print('DIRECTORY TO SYNC: {}'.format(directory))
    print('\n')
    if directory != '.':
        print('CHECKING IF DIRECTORY {} IN: {}'.format(
            directory.split('/')[-1], rootdir))
    check_dir = devIO.dev.wr_cmd("os.listdir('{}')".format(rootdir), silent=True,
                                 rtn_resp=True)
    # print(check_dir)
    if directory.split('/')[-1] in check_dir or directory == '.':
        print('DIRECTORY {} FOUND'.format(directory))
        print('\n')
        print('FILES/DIRS IN DIRECTORY {}:'.format(directory))
        if directory == '.':
            directory = ''
        _directory = directory if directory else '.'
        devIO.dev.wr_cmd("du(path='{}', hidden=True)".format(_directory), follow=True)
        # for file in os.listdir(directory):
        #     print('- {} {} [{}]'.format(type_file_dict[os.path.isfile(os.path.join(current_dir, file))],
        #                            file, print_filesys_info(os.stat(os.path.join(current_dir, file))[6])))
        if directory == '':
            directory = '.'
        file_list = devIO.dev.wr_cmd("os.listdir('{}')".format(directory), silent=True,
                                     rtn_resp=True)
        print('\n')
        # if args.wdl:
        #     modified_files, deleted_files = check_wdlog(path=directory)
        is_dir_cmd = "import os; os.stat('{}')[0] & 0x4000"
        for file in file_list:
            if not devIO.dev.wr_cmd(is_dir_cmd.format(os.path.join(current_dir, file)),
                                    silent=True, rtn_resp=True):  # os.path.join(current_dir, file)
                # if args.wdl:
                #     wdl_file = file.split('/')[-1]
                #     if wdl_file not in _CONFG_FILES and wdl_file in modified_files:
                #         file_list_abs_path.append(os.path.join(current_dir, file))
                # else:
                file_list_abs_path.append(os.path.join(current_dir, file))
            else:
                dir_list_abs_path.append(os.path.join(current_dir, file))

    if file_list_abs_path:
        print('LIST OF FILES TO DOWNLOAD:')
        for file in file_list_abs_path:
            print('- {}'.format(file.split('/')[-1]))
        print('\n')
    else:
        pass
        # print('NO FILES TO UPLOAD')

    if dir_list_abs_path:
        print('LIST OF SUBDIRS TO CREATE:')
        for subdir in dir_list_abs_path:
            print('- {}'.format(subdir.split('/')[-1]))
        print('\n')
    else:
        # print('NO SUBDIRS TO CREATE')
        pass
    # Now create the root sync dir:
    if './' == rootdir:
        rootdir = '.'
    # if not devIO.dev.connected:
    #     devIO.dev.connect()
    host_root_list = current_dir.split('/')[-1] in os.listdir(rootdir)
    if not host_root_list and current_dir != '.':
        print('\n')
        print('MAKING DIR: {}'.format(current_dir))
        print('\n')
        # devIO.dev.wr_cmd("os.mkdir('{}')".format(current_dir), silent=True)
        os.mkdir(current_dir)
        print('\n')

    _files_to_delete_in_host = []
    if file_list_abs_path:
        print('DOWNLOADING FILES TO {}'.format(current_dir))
        os.chdir(current_dir)
        if len(file_list_abs_path) > 1:
            if directory == '.':
                file_list_abs_path = [file.split('/')[-1] for file in file_list_abs_path]
            for file in file_list_abs_path:
                print('- {}'.format(file))
            args.fre = file_list_abs_path
            if directory != '.':
                args.s = os.path.join(*file_list_abs_path[0].split('/')[:-1])
            else:
                args.s = '/'
            print('\n')
            if args.rf:
                # print(args.fre)
                _files_to_delete_in_host += [os.path.join(current_dir, _file) for _file in os.listdir() if os.path.isfile(_file) and os.path.join(current_dir, _file) not in file_list_abs_path]
                # print(_files_to_delete_in_host)
            devIO.get_files(args, dev_name)
            args.fre = None
            time.sleep(0.2)
        elif len(file_list_abs_path) == 1:
            args.fre = None
            file_to_get = file_list_abs_path[0]
            if directory == '.':
                file_to_get = file.split('/')[-1]
            print('- {}'.format(file_to_get), end='\n\n')
            file_to_get_from_dev = file_to_get.replace('./', '')
            devIO.get(file_to_get, file_to_get_from_dev, ppath=True, dev_name=dev_name)
            time.sleep(0.2)
            if args.rf:
                # print(file_to_get, file_to_get_from_dev)
                _files_to_delete_in_host += [os.path.join(current_dir, _file) for _file in os.listdir() if os.path.isfile(_file) and os.path.join(current_dir, _file) != file_to_get]
                # print(_files_to_delete_in_host)
        if current_dir != '.':
            for level in current_dir.split('/'):
                if level != '.':
                    os.chdir('..')
    else:
        print('NO FILES TO DOWNLOAD')
        # if args.rf check files to delete
        if args.rf:
            _files_to_delete_in_host += [os.path.join(current_dir, _file) for _file in os.listdir(current_dir) if os.path.isfile(os.path.join(current_dir, _file))]
            # print(_files_to_delete_in_host)
    # Now create subdirs:
    print('\n')
    _dirs_to_delete_in_host = []
    if dir_list_abs_path:
        print('MAKING DIRS NOW...')
        if args.rf:
            # print(current_dir, os.listdir(current_dir))
            if current_dir in os.listdir():
                _dirs_to_delete_in_host += [os.path.join(current_dir, _dir) for _dir in os.listdir(current_dir) if os.path.isdir(os.path.join(current_dir, _dir)) and os.path.join(current_dir, _dir) not in dir_list_abs_path]
            # print(_dirs_to_delete_in_host)
        for dir_ in dir_list_abs_path:
            print('\n')
            current_dir = dir_
            # Now create the root sync dir:
            try:
                dir_to_sync = directory
                dir_to_find = current_dir.split('/')[-1]
                # print(os.getcwd())
                dev_directory = dir_to_find in os.listdir(dir_to_sync)
            except Exception as e:
                print(e, dir_to_sync)
            if not dev_directory:
                print('Creating dir: {}'.format(current_dir))
                os.mkdir(current_dir)
            else:
                print('DIRECTORY {} ALREADY EXISTS'.format(dir_to_find))
    else:
        print('NO DIRS TO MAKE')
        # if args.rf check dirs to delete
        # print(current_dir, os.listdir(current_dir))
        if args.rf:
            _dirs_to_delete_in_host += [os.path.join(current_dir, _dir) for _dir in os.listdir(current_dir) if os.path.isdir(os.path.join(current_dir, _dir))]
        # print(_dirs_to_delete_in_host)

    if args.rf:
        # Delete Files

        if _files_to_delete_in_host:
            print('FILES TO REMOVE:')
            for dfile in _files_to_delete_in_host:
                try:
                    os.remove(dfile)
                    if directory != '.':
                        print('- {}'.format(dfile))
                    else:
                        print('- {}'.format(dfile))
                except Exception as e:
                    print(e)
            print('\nFILES DELETED')

        # Deleted Directories

        if _dirs_to_delete_in_host:

            print('\nDIRS TO REMOVE:')
            for ddir in _dirs_to_delete_in_host:
                print('- {} '.format(ddir))
                is_dir_empty = not len(os.listdir(ddir)) > 0
                if is_dir_empty:
                    os.rmdir(ddir)
                else:
                    rmrf(ddir)
            print('\nDIRS DELETED')


    root = directory
    for dir_ in dir_list_abs_path:
        dev2host_sync_recursive(dir_, devIO, root, args=args, dev_name=dev_name)

    if directory == root_sync_folder:
        print('Done in : {:.2f} seconds'.format(time.time()-t0))



from upydevice import Device
import sys
from upydev.helpinfo import see_help
from upydevice import check_device_type
import os
from datetime import datetime, timedelta
import requests
import shlex
import subprocess
import upydev
import re
import time

UPYDEV_PATH = upydev.__path__[0]

FIRMWARE_HELP = """
> FIRMWARE: Usage '$ upydev ACTION [opts]'
    ACTIONS:
        - fwr: to list or get available firmware versions, use -md option to indicate operation:
                to list do: "upydev fwr -md list -b [BOARD]" board can be e.g. 'esp32','esp8266' or 'PYBD'
                            "upydev fwr -md list latest -b [BOARD]" to see the latest firmware available
                to get do: "upydev fwr -md get [firmware file]" or "upydev fwr -md get latest -b[BOARD]"
                * for list or get modes the -n option will filter the results further: e.g. -n ota
                to see available serial ports do: "upydev fwr -md list serial_ports"

        - flash: to flash a firmware file to the upydevice, a serial port must be indicated
                    to flash do: "upydev flash -port [serial port] -f [firmware file]"

        - mpyx : To froze a module/script , and save some RAM, it uses mpy-cross tool
                 (mpy-cross must be available in $PATH) e.g. $ upydev mpyx [FILE].py,
                 $ upydev mpyx [FILE].py [FILE2].py, $ upydev mpyx *.py"""


def get_fw_versions(keyword):
    fw_list = []
    fw_links = []
    r = requests.get('https://micropython.org/download/all/')
    fw_text = [line for line in r.text.split(
        '\n') if keyword in line and any(x in line for x in ['bin', 'dfu', 'zip']) and 'firmware' in line]
    if not fw_text:
        keyword = keyword[:3]
        fw_text = [line for line in r.text.split(
            '\n') if keyword in line and any(x in line for x in ['bin', 'dfu', 'zip']) and 'firmware' in line]
    for line in fw_text:
        for element in re.split(r"[=<>]+", line):
            if 'firmware' in element:
                fw_links.append('www.micropython.org{}'.format(element[1:-1]))
                fw_list.append(element[1:-1].split('/')[3])
    if not fw_text:
        keyword = keyword[:3]
    fw_dict = dict(zip(fw_list, fw_links))
    return fw_dict, fw_list


def firmwaretools_action(args, **kargs):

    # MPYX

    if args.m == 'mpyx':
        if not args.f:
            print('File name required indicate with -f option.')
            see_help(args.m)
        else:
            if args.f not in os.listdir():
                print('File not found, indicate a valid file')
            else:
                print('Frozing {} to bytecode with mpy-cross'.format(args.f))
                mpy_cmd_str = 'python -m mpy_cross {}'.format(args.f)
                mpy_cmd = shlex.split(mpy_cmd_str)
                try:
                    mpy_tool = subprocess.call(mpy_cmd)
                    if mpy_tool == 0:
                        print('Process successful, bytecode in : {}'.format(
                                                args.f.replace('py', 'mpy')))
                    else:
                        print('Process failed.')
                except KeyboardInterrupt:
                    pass

        sys.exit()

    # FW
    elif args.m == 'fwr':
        if not args.md:
            print('Argument mode required (list, or get), indicate with -md')
            see_help(args.m)
            sys.exit()
        else:
            if args.md[0] == 'list':
                if not args.b:  # Autodetect
                    try:
                        dev = Device(args.t, args.p, init=True, autodetect=True,
                                     ssl=args.wss, auth=args.wss)
                        args.b = dev.dev_platform
                        if dev.dev_platform == 'pyboard':
                            machine = dev.cmd('import os;os.uname().machine',
                                              silent=True, rtn_resp=True)
                            machine, version = machine.split()[0].split('v')
                            version_str = version.replace('.', '')
                            machine = machine.lower()
                            args.b = "{}v{}".format(machine, version_str)
                        dev.disconnect()
                    except Exception as e:
                        print(e)
                        print('Indicate a device platform with -b option')
                        sys.exit()
                if len(args.md) > 1:
                    if args.md[1] == 'latest':
                        today = datetime.strftime(datetime.now(), "%Y%m%d")
                        fw_v = get_fw_versions(args.b)
                        if fw_v[1]:
                            print('Latest Firmware versions found for {}:'.format(args.b))
                            fw_v_latest = [v for v in fw_v[1] if today in v]
                            days_before = 0
                            while len(fw_v_latest) == 0:
                                if len(fw_v[1]) == 0:
                                    break
                                days_before += 1
                                latest_fw = datetime.strftime(
                                    datetime.now()-timedelta(days=days_before), "%Y%m%d")
                                fw_v_latest = [v for v in fw_v[1] if latest_fw in v]
                            if args.n is not None:
                                if args.n != 'def':
                                    fw_v_latest_opt = [
                                        v for v in fw_v_latest if args.n in v]
                                    for v in fw_v_latest_opt:
                                        print('- {}'.format(v))
                                else:
                                    print('- {}'.format(fw_v_latest[0]))
                            else:
                                for v in fw_v_latest:
                                    print('- {}'.format(v))
                        else:
                            print('No firmware available that match: {}'.format(args.b))
                        # fw_v_latest_link = fw_v[0][fw_v_latest]

                else:
                    fw_v = get_fw_versions(args.b)[1]
                    if not fw_v:
                        print('No firmware available that match: {}'.format(args.b))
                    else:
                        if args.n is not None:
                            print('Firmware versions found for {}-{}:'.format(args.b, args.n))
                            if args.n != 'def':
                                fw_v_opt = [v for v in fw_v if args.n in v]
                                for v in fw_v_opt:
                                    print('- {}'.format(v))
                            else:
                                print('- {}'.format(fw_v[0]))
                        else:
                            print('Firmware versions found for {}:'.format(args.b))
                            for version in fw_v:
                                print('- {}'.format(version))
            elif args.md[0] == 'get':
                if len(args.md) > 1:
                    if args.md[1] == 'latest':
                        if not args.b:  # Autodetect
                            try:
                                dev = Device(args.t, args.p, init=True, autodetect=True,
                                             ssl=args.wss, auth=args.wss)
                                args.b = dev.dev_platform
                                if dev.dev_platform == 'pyboard':
                                    machine = dev.cmd('import os;os.uname().machine',
                                                      silent=True, rtn_resp=True)
                                    machine, version = machine.split()[0].split('v')
                                    version_str = version.replace('.', '')
                                    machine = machine.lower()
                                    args.b = "{}v{}".format(machine, version_str)
                                dev.disconnect()
                            except Exception as e:
                                print(e)
                                print('Indicate a device platform with -b option')
                                sys.exit()
                        today = datetime.strftime(datetime.now(), "%Y%m%d")
                        fw_v = get_fw_versions(args.b)

                        if fw_v[1]:
                            print('Latest version found for {}:'.format(args.b))
                            fw_v_latest = [v for v in fw_v[1] if today in v]
                            days_before = 0
                            while len(fw_v_latest) == 0:
                                if len(fw_v[1]) == 0:
                                    break
                                days_before += 1
                                latest_fw = datetime.strftime(
                                    datetime.now()-timedelta(days=days_before), "%Y%m%d")
                                fw_v_latest = [v for v in fw_v[1] if latest_fw in v]
                            if args.n is not None:
                                if args.n != 'def':
                                    fw_v_latest = [
                                        v for v in fw_v_latest if args.n in v]
                                    for v in fw_v_latest:
                                        print('- {}'.format(v))
                            else:
                                print('- {}'.format(fw_v_latest[0]))
                            if len(fw_v_latest) >= 1:
                                fw_v_latest = fw_v_latest[0]
                                print('Firmware selected: {}'.format(fw_v_latest))
                                fw_v_latest_link = fw_v[0][fw_v_latest]
                                print('Downloading {} ...'.format(fw_v_latest))
                                curl_cmd_str = "curl -O '{}'".format(fw_v_latest_link)
                                curl_cmd = shlex.split(curl_cmd_str)
                                try:
                                    proc = subprocess.call(curl_cmd)
                                    print('Done!')
                                except KeyboardInterrupt:
                                    print('Operation cancelled')
                        else:
                            print('No firmware available that match: {}'.format(args.b))
                    else:
                        fw_v_link = get_fw_versions(args.md[1])[0][args.md[1]]
                        print('Downloading {} ...'.format(args.md[1]))
                        curl_cmd_str = "curl -O '{}'".format(fw_v_link)
                        curl_cmd = shlex.split(curl_cmd_str)
                        try:
                            proc = subprocess.call(curl_cmd)
                            print('Done!')
                        except KeyboardInterrupt:
                            print('Operation cancelled')
                else:
                    if not args.b:  # Autodetect
                        try:
                            dev = Device(args.t, args.p, init=True, autodetect=True,
                                         ssl=args.wss, auth=args.wss)
                            args.b = dev.dev_platform
                            if dev.dev_platform == 'pyboard':
                                machine = dev.cmd('import os;os.uname().machine',
                                                  silent=True, rtn_resp=True)
                                machine, version = machine.split()[0].split('v')
                                version_str = version.replace('.', '')
                                machine = machine.lower()
                                args.b = "{}v{}".format(machine, version_str)
                            dev.disconnect()
                        except Exception as e:
                            print(e)
                            print('Indicate a device platform with -b option')
                            sys.exit()
                    fw_v_dict, fw_v = get_fw_versions(args.b)
                    if not fw_v:
                        print('No firmware available that match: {}'.format(args.b))
                    else:
                        fw_v_opt = []
                        if args.n is not None:
                            print('Firmware versions found for {}-{}:'.format(args.b, args.n))
                            if args.n != 'def':
                                fw_v_opt = [v for v in fw_v if args.n in v]
                                for v in fw_v_opt:
                                    print('- {}'.format(v))
                            else:
                                print('- {}'.format(fw_v[0]))
                        else:
                            print('Firmware versions found for {}:'.format(args.b))
                            for version in fw_v:
                                print('- {}'.format(version))
                        if fw_v_opt:
                            fw_v = fw_v_opt[0]
                        else:
                            fw_v = fw_v[0]
                        print('Firmware selected: {}'.format(fw_v))
                        fw_v_link = fw_v_dict[fw_v]
                        print('Downloading {} ...'.format(fw_v))
                        curl_cmd_str = "curl -O '{}'".format(fw_v_link)
                        curl_cmd = shlex.split(curl_cmd_str)
                        try:
                            proc = subprocess.call(curl_cmd)
                            print('Done!')
                        except KeyboardInterrupt:
                            print('Operation cancelled')

    elif args.m == 'flash':
        devname = kargs.get('device')
        if not args.f:
            print('Firmware file name required, indicate with -f option')
            see_help(args.m)
            sys.exit()
        else:
            dt = check_device_type(args.t)
            if not args.port and dt != 'SerialDevice':
                print(
                    'Indicate a serial port with -port option, {} is NOT a SerialDevice'.format(devname))
                sys.exit()
            else:
                if not args.port:
                    args.port = args.t
                if 'esp32' in args.f:
                    if args.i:
                        print('Checking firmware and device platform match')
                        try:
                            dev = Device(args.t, args.p, init=True, autodetect=True,
                                         ssl=args.wss, auth=args.wss)
                            platform = dev.dev_platform
                            dev.disconnect()
                        except Exception as e:
                            print(e)
                            print('Device not reachable, connect the device and try again.')
                            sys.exit()
                        if platform in args.f:
                            print(
                                'Firmware {} and {} device platform [{}] match, flashing firmware now...'.format(args.f, devname, platform))
                        else:
                            print('Firmware {} and {} device platform [{}] do NOT match, operation aborted.'.format(args.f, devname, platform))
                            sys.exit()
                    print('Flashing firmware {} with esptool.py to {} @ {}...'.format(
                        args.f, devname, args.port))
                    esptool_cmd_str = "esptool.py --chip esp32 --port {} write_flash -z 0x1000 {}".format(
                        args.port, args.f)
                    print(esptool_cmd_str)
                    esptool_cmd = shlex.split(esptool_cmd_str)
                    try:
                        proc = subprocess.call(esptool_cmd)
                        print('Done!')
                    except KeyboardInterrupt:
                        print('Operation cancelled')
                elif 'esp8266' in args.f:
                    if args.i:
                        print('Checking firmware and device platform match')
                        try:
                            dev = Device(args.t, args.p, init=True, autodetect=True,
                                         ssl=args.wss, auth=args.wss)
                            platform = dev.dev_platform
                            dev.disconnect()
                        except Exception as e:
                            print(e)
                            print('Device not reachable, connect the device and try again.')
                            sys.exit()
                        if platform in args.f:
                            print(
                                'Firmware {} and {} device platform [{}] match, flashing firmware now...'.format(args.f, devname, platform))
                        else:
                            print('Firmware {} and {} device platform [{}] do NOT match, operation aborted.'.format(args.f, devname, platform))
                            sys.exit()
                    print('Flashing firmware {} with esptool.py to {} @ {}...'.format(
                        args.f, devname, args.port))
                    esptool_cmd_str = "esptool.py --port {} --baud 460800 write_flash --flash_size=detect 0 {}".format(
                        args.port, args.f)
                    print(esptool_cmd_str)
                    esptool_cmd = shlex.split(esptool_cmd_str)
                    try:
                        proc = subprocess.call(esptool_cmd)
                        print('Done!')
                    except KeyboardInterrupt:
                        print('Operation cancelled')

                elif 'pyb' in args.f:
                    if args.i:
                        print('Checking firmware and device platform match')
                        try:
                            dev = Device(args.t, args.p, init=True, autodetect=True,
                                         ssl=args.wss, auth=args.wss)
                            machine = dev.cmd('import os;os.uname().machine', silent=True,
                                              rtn_resp=True)
                            dev.disconnect()
                        except Exception as e:
                            print(e)
                            print('Device not reachable, connect the device and try again.')
                            sys.exit()
                        platform = dev.dev_platform
                        if platform in args.f or platform[:3] in args.f:
                            machine, version = machine.split()[0].split('v')
                            version_str = version.replace('.', '')
                            machine = machine.lower()
                            fw_string = "{}v{}".format(machine, version_str)

                            if fw_string == args.f.split('-')[0]:
                                print(
                                    'Firmware {} and {} device platform [{}] machine [{}], verison [{}] match, flashing firmware now...'.format(args.f, devname, platform, machine, version))
                            else:
                                print('Firmware {} and {} device platform [{}] machine [{}], verison [{}] do NOT match, operation aborted.'.format(args.f, devname, platform, machine, version))
                                sys.exit()
                        else:
                            print('Firmware {} and {} device platform [{}] do NOT match, operation aborted.'.format(args.f, devname, platform))
                            sys.exit()

                    print('Enabling DFU mode in pyboard, DO NOT DISCONNECT...')
                    dev.connect()
                    bs = dev.serial.write(bytes('import pyb;pyb.bootloader()\r', 'utf-8'))
                    time.sleep(0.2)
                    bin_file = args.f
                    flash_tool = input('Select a tool to flash pydfu.py/dfu-util: (0/1) ')
                    if flash_tool == '0':
                        print('Using pydfu.py...')
                        pydfu_fw_cmd = 'pydfu -u {}'.format(bin_file)
                        # print(upy_fw_cmd)
                        flash_fw_cmd = shlex.split(pydfu_fw_cmd)

                        try:
                            fw_flash = subprocess.call(flash_fw_cmd)
                        except Exception as e:
                            # shr_cp.enable_wrepl_io()
                            print(e)
                        print('Flashing firmware finished successfully!')
                        time.sleep(1)
                        print('Done!')
                    else:
                        print('Using dfu-util...')
                        dfu_fw_cmd = 'sudo dfu-util --alt 0 -D {}'.format(bin_file)
                        # print(upy_fw_cmd)
                        flash_fw_cmd = shlex.split(dfu_fw_cmd)

                        try:
                            fw_flash = subprocess.call(flash_fw_cmd)
                        except Exception as e:
                            # shr_cp.enable_wrepl_io()
                            print(e)
                        print('Flashing firmware finished successfully, RESET the pyboard now.')
                        time.sleep(1)

        sys.exit()
